from dataclasses import replace
from typing import Any, Optional, Tuple, Union
from typing_extensions import override

from PIL import Image
import numpy as np
import torch
from torch import Tensor

from conduit.data.datasets.utils import extract_base_dataset
from conduit.data.datasets.vision.base import CdtVisionDataset
from conduit.data.datasets.vision.utils import ImageTform, apply_image_transform
from conduit.data.structures import Dataset, DatasetWrapper, RawImage, SampleBase

__all__ = [
    "ImageTransformer",
]


class ImageTransformer(DatasetWrapper[Any]):
    """
    Wrapper class for applying image transformations.

    Useful when wanting to have different transformations for different subsets of the data
    that share the same underlying dataset.
    """

    def __init__(self, dataset: Dataset, *, transform: Optional[ImageTform]) -> None:
        self.dataset = dataset
        self._transform: Optional[ImageTform] = None
        self.transform = transform

    @property
    def transform(self) -> Optional[ImageTform]:
        return self._transform

    @transform.setter
    def transform(self, transform: Optional[ImageTform]) -> None:
        base_dataset = extract_base_dataset(self.dataset, return_subset_indices=False)
        if isinstance(base_dataset, CdtVisionDataset):
            base_dataset.update_il_backend(transform)
        self._transform = transform

    @override
    def __getitem__(self, index: int) -> Any:
        sample = self.dataset[index]

        def _transform_sample(
            _sample: Union[RawImage, SampleBase, Tuple[Union[RawImage, Image.Image], ...]],
        ) -> Any:
            if self.transform is None:
                return _sample
            if isinstance(_sample, (Image.Image, np.ndarray)):
                return apply_image_transform(image=_sample, transform=self.transform)
            elif isinstance(_sample, SampleBase):
                image = _sample.x
                if isinstance(image, list):
                    image = [_transform_sample(elem) for elem in image]
                    if isinstance(image[0], Tensor):
                        image = torch.stack(image, dim=0)
                    elif isinstance(image[0], np.ndarray):
                        image = np.stack(image, axis=0)

                elif not isinstance(image, (Image.Image, np.ndarray)):
                    raise TypeError(
                        f"Image transform cannot be applied to input of type '{type(image)}'"
                        "(must be a PIL Image or a numpy array)."
                    )
                else:
                    image = apply_image_transform(image=image, transform=self.transform)
                return replace(_sample, x=image)
            else:
                image = apply_image_transform(image=_sample[0], transform=self.transform)
                data_type = type(_sample)
                return data_type(image, *_sample[1:])  # type: ignore

        return _transform_sample(sample)



from pathlib import Path
from typing import ClassVar, Optional, Union
from typing_extensions import TypeAlias

import pandas as pd
from ranzen import StrEnum
import torch
from torch import Tensor

from conduit.data.datasets.utils import UrlFileInfo, download_from_url
from conduit.data.structures import SubgroupSample, TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["IWildCam", "IWildCamSplit", "IWildCamUnlabeled"]


class IWildCamSplit(StrEnum):
    TRAIN = "train"
    VAL = "val"
    ID_VAL = "id_val"
    TEST = "test"
    ID_TEST = "id_test"


class IWildCam(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """The iWildCam2020-WILDS dataset.
    This is a port of the iWildCam dataset from `WILDS`_ which itself is a  modified version of the
    original iWildCam2020 competition dataset introduced in `iWildCam`_.

    Per `iWildCam`_: Camera traps enable the automatic collection of large quantities of image data.
    Biologists all over the world use camera traps to monitor animal populations. We have recently
    been making strides towards automatic species classification ['y' label] in camera trap images
    ['s' label]. However, as we try to expand the geographic scope of these models we are faced with
    an interesting question: how do we train models that perform well on new (unseen during
    training) camera trap locations? Can we leverage data from other modalities, such as citizen
    science data and remote sensing data? In order to tackle this problem, we have prepared a
    challenge where the training data and test data are from different cameras spread across the
    globe. For each camera, we provide a series of remote sensing imagery that is tied to the
    location of the camera. We also provide citizen science imagery from the set of species seen in
    our data. The challenge is to correctly classify species ('y') in the test camera traps.

    .. _WILDS:
        https://arxiv.org/abs/2012.07421

    .. _iWildCam:
        https://arxiv.org/abs/2004.10340
    """

    SampleType: TypeAlias = TernarySample
    Split: TypeAlias = IWildCamSplit

    _BASE_DIR_NAME: ClassVar[str] = "iwildcam_v2.0"
    _IMG_DIR_NAME: ClassVar[str] = "train"
    _METADATA_FILENAME: ClassVar[str] = "metadata.csv"

    _FILE_INFO: ClassVar[UrlFileInfo] = UrlFileInfo(
        name="iwildcam.tar.gz",
        url="https://worksheets.codalab.org/rest/bundles/0xff56ea50fbf64aabbc4d09b2e8d50e18/contents/blob/",
        md5=None,
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        split: Optional[Union[IWildCamSplit, str]] = None,
    ) -> None:
        """
        :param root: Root directory of the dataset.
        :param split: Which predefined split of the dataset to use. If ``None`` then the full
            (unsplit) dataset will be returned.
        :param transform: A function/transform that takes in a PIL or ndarray image and returns a
            transformed version. E.g, ``transforms.RandomCrop``.
        :param download: If ``True``, downloads the dataset from the internet and puts it in the
            root directory. If the dataset is already downloaded, it is not downloaded again.
        :raises FileNotFoundError: If ``download=False`` and an existing dataset cannot be found in
            the root directory.
        """
        self.split = IWildCamSplit(split) if isinstance(split, str) else split
        self.root = Path(root)
        self._base_dir = self.root / self._BASE_DIR_NAME
        self._img_dir = self._base_dir / self._IMG_DIR_NAME
        self.download = download
        if not self._check_unzipped():
            if self.download:
                download_from_url(
                    file_info=self._FILE_INFO,
                    root=self.root,
                    logger=self.logger,
                    remove_finished=True,
                )

            else:
                raise FileNotFoundError(
                    f"Data not found at location {self._base_dir.resolve()}. Have you"
                    " downloaded it?"
                )

        # Read in metadata
        self.metadata = pd.read_csv(self._base_dir / self._METADATA_FILENAME)
        # Use an official split of the data, if specified, else just use all of the data
        if self.split is not None:
            split_indices = self.metadata["split"] == self.split.value
            self.metadata = self.metadata.loc[split_indices]

        # Extract filenames
        x = self.metadata['filename'].to_numpy()
        # Extract class (species) labels
        y = torch.as_tensor(self.metadata["y"].to_numpy(), dtype=torch.long)
        # Extract camera-trap-location labels
        s = torch.as_tensor(self.metadata["location_remapped"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._img_dir)

    def _check_unzipped(self) -> bool:
        return (self._img_dir).is_dir()


class IWildCamUnlabeled(CdtVisionDataset[SubgroupSample, None, Tensor]):
    """An extra unlablled (with respect to 'y') split of the WILDS, introduced in `Ext_WILDS`_.

    `_Ext_WILDS`_ extends the WILDS benchmark to include curated unlabeled data that would be
    realistically obtainable in deployment. In the case of the iWildCam dataset specifically, the
    unlabeled data, comprising 819,120 images, is obtained from a set of WCS camera traps entirely
    disjoint with the labeled dataset, representative of unlabeled data from a newly-deployed sensor
    network.

    .. _Ext_WILDS:
        https://arxiv.org/abs/2012.07421
    """

    SampleType: TypeAlias = SubgroupSample
    Split: TypeAlias = IWildCamSplit

    _BASE_DIR_NAME: ClassVar[str] = "iwildcam_unlabeled_v1.0"
    _IMG_DIR_NAME: ClassVar[str] = "images"
    _FILE_INFO: ClassVar[UrlFileInfo] = UrlFileInfo(
        name="iwildcam_unlabeled.tar.gz",
        url="https://worksheets.codalab.org/rest/bundles/0x6313da2b204647e79a14b468131fcd64/contents/blob/",
        md5=None,
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
    ) -> None:
        self.root = Path(root)
        self._base_dir = self.root / self._BASE_DIR_NAME
        self._img_dir = self._base_dir / self._IMG_DIR_NAME
        self.download = download
        if self.download:
            download_from_url(
                file_info=self._FILE_INFO,
                root=self.root,
                logger=self.logger,
                remove_finished=True,
            )

        elif not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )

        # Read in metadata
        self.metadata = pd.read_csv(self._base_dir / 'metadata.csv')
        self.metadata["filename"] = self.metadata["uid"] + ".jpg"

        # Extract filenames
        x = self.metadata['filename'].to_numpy()
        # Extract camera-trap-location labels
        s = torch.as_tensor(self.metadata["location_remapped"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=None, s=s, transform=transform, image_dir=self._img_dir)

    def _check_unzipped(self) -> bool:
        return (self._img_dir).is_dir()



"""Data structures."""

from abc import abstractmethod
from collections.abc import Sequence
from dataclasses import asdict, dataclass, field, fields, is_dataclass
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)
from typing_extensions import Self, TypeAlias, override

from PIL import Image
import numpy as np
import numpy.typing as npt
from ranzen.misc import gcopy, reduce_add
from ranzen.types import Addable
import torch
from torch import Tensor

from conduit.types import IndexType, Sized

__all__ = [
    "BinarySample",
    "BinarySampleIW",
    "Dataset",
    "DatasetWrapper",
    "ImageSize",
    "InputContainer",
    "LoadedData",
    "MeanStd",
    "MultiCropOutput",
    "NamedSample",
    "PseudoCdtDataset",
    "RawImage",
    "SampleBase",
    "SizedDataset",
    "SubgroupSample",
    "SubgroupSampleIW",
    "TargetData",
    "TernarySample",
    "TernarySampleIW",
    "TrainTestSplit",
    "TrainValTestSplit",
    "UnloadedData",
    "concatenate_inputs",
    "shallow_asdict",
    "shallow_astuple",
]


@runtime_checkable
class InputContainer(Sized, Addable, Protocol):
    @classmethod
    def fromiter(cls, sequence: Iterable[Self]) -> Self:
        """
        Collates a sequence of container instances into a single instance.

        :param sequence: Sequence of containers to be collated.

        :returns: A collated container.
        """
        return reduce_add(sequence)

    @override
    def __len__(self) -> int:
        """Total number of samples in the container."""
        ...

    @override
    def __add__(self, other: Self) -> Self:
        ...

    def to(
        self,
        device: Optional[Union[torch.device, str]],
        *,
        non_blocking: bool = False,
    ) -> Self:
        for name, value in shallow_asdict(self).items():
            if isinstance(value, (Tensor, InputContainer)):
                setattr(self, name, value.to(device, non_blocking=non_blocking))
        return self


RawImage: TypeAlias = Union[npt.NDArray[np.integer], Image.Image]
UnloadedData: TypeAlias = Union[
    npt.NDArray[np.floating],
    npt.NDArray[np.integer],
    npt.NDArray[np.bytes_],
    Tensor,
]
LoadedData: TypeAlias = Union[
    Tensor,
    Image.Image,
    npt.NDArray[np.floating],
    npt.NDArray[np.integer],
    npt.NDArray[np.bytes_],
    Dict[str, Tensor],
    Dict[str, Image.Image],
    Dict[str, npt.NDArray[np.floating]],
    Dict[str, npt.NDArray[np.integer]],
    Dict[str, npt.NDArray[np.bytes_]],
    List[Image.Image],
    InputContainer,
]
IndexabledData: TypeAlias = Union[
    Tensor,
    npt.NDArray[np.floating],
    npt.NDArray[np.integer],
    npt.NDArray[np.bytes_],
]

X = TypeVar("X", bound=LoadedData)
X_co = TypeVar("X_co", bound=LoadedData, covariant=True)
XI = TypeVar("XI", bound=IndexabledData)

TargetData: TypeAlias = Union[Tensor, npt.NDArray[np.floating], npt.NDArray[np.integer]]


def concatenate_inputs(x1: X, x2: X, *, is_batched: bool) -> X:
    if type(x1) != type(x2) or (isinstance(x1, list) and type(x1[0]) != type(cast(List, x2)[0])):
        raise AttributeError("Only data of the same type can be concatenated (added) together.")
    if isinstance(x1, Tensor):
        # if the number of dimensions is different by 1, append a batch dimension.
        ndim_diff = x1.ndim - x2.ndim  # type: ignore
        if ndim_diff == 1:
            x2 = x2.unsqueeze(0)  # type: ignore
        elif ndim_diff == -1:
            x1 = x1.unsqueeze(0)
        if is_batched:
            return torch.cat([x1, x2], dim=0)  # type: ignore
        return torch.stack([x1, x2], dim=0)  # type: ignore

    elif isinstance(x1, np.ndarray):
        # if the number of dimensions is different by 1, append a batch dimension.
        ndim_diff = x1.ndim - x2.ndim  # type: ignore
        if ndim_diff == 1:
            x2 = np.expand_dims(x2, axis=0)  # type: ignore
        elif ndim_diff == -1:
            x1 = np.expand_dims(x1, axis=0)  # type: ignore
        if is_batched:
            return np.concatenate([x1, x2], axis=0)  # type: ignore
        return np.stack([x1, x2], axis=0)  # type: ignore
    elif isinstance(x1, Image.Image):
        return [x1, x2]  # type: ignore
    elif isinstance(x1, dict):
        for key, value in x2.items():  # type: ignore
            if key in x1:
                x1[key] = concatenate_inputs(x1[key], value, is_batched=is_batched)  # type: ignore
            else:
                x1[key] = value  # type: ignore
            return x1
    return x1 + x2  # type: ignore


@dataclass
class MultiCropOutput(InputContainer):
    global_crops: List[Tensor]
    local_crops: List[Tensor] = field(default_factory=list)

    @property
    def all_crops(self) -> List[Tensor]:
        return self.global_crops + self.local_crops

    @property
    def global_crop_sizes(self) -> List[torch.Size]:
        return [crop.shape[-3:] for crop in self.global_crops]

    @property
    def local_crop_sizes(self) -> List[torch.Size]:
        return [crop.shape[-3:] for crop in self.local_crops]

    @property
    def shape(self) -> torch.Size:
        """Shape of the global crops - for compatibility with DMs."""
        return self.global_crops[0].shape

    @override
    def __len__(self) -> int:
        """Total number of crops."""
        return len(self.global_crops) + len(self.local_crops)

    def __iadd__(self, other: Self) -> Self:
        self.global_crops += other.global_crops
        self.local_crops += other.local_crops
        return self

    @override
    def __add__(self, other: Self) -> Self:
        copy = gcopy(self, deep=False)
        copy.global_crops = copy.global_crops + other.global_crops
        copy.local_crops = copy.local_crops + other.local_crops
        return copy


@dataclass
class SampleBase(InputContainer, Generic[X]):
    x: X

    @override
    def __len__(self) -> int:
        return len(self.__dataclass_fields__)

    @abstractmethod
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        ...

    @override
    def __add__(self, other: Self) -> Self:
        return self._get_copy(other, is_batched=True)

    def _get_copy(self, other: Self, is_batched: bool) -> Self:
        copy = gcopy(self, deep=False)
        copy.x = concatenate_inputs(copy.x, other.x, is_batched=is_batched)
        return copy

    def astuple(self, deep: bool = False) -> Tuple[Union[X, Tensor], ...]:
        tuple_ = tuple(iter(self))
        if deep:
            tuple_ = gcopy(tuple_, deep=True)
        return tuple_

    def asdict(self, deep: bool = False) -> Dict[str, X]:
        if deep:
            asdict(self)
        return shallow_asdict(self)

    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index])


@dataclass
class _BinarySampleMixin:
    y: Tensor

    def _add_to_y(self, other: Self) -> None:
        self.y = torch.cat([self.y, other.y], dim=0)


@dataclass
class _SubgroupSampleMixin:
    s: Tensor

    def _add_to_s(self, other: Self) -> None:
        self.s = torch.cat([self.s, other.s], dim=0)


@dataclass
class _IwMixin:
    iw: Tensor

    def _add_to_iw(self, other: Self) -> None:
        self.iw = torch.cat([self.iw, other.iw], dim=0)


@dataclass
class TernarySampleIW(_IwMixin, _BinarySampleMixin, _SubgroupSampleMixin, SampleBase[X]):
    def add_field(self) -> Self:
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.y, self.s, self.iw)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.y) > 1)
        copy._add_to_y(other)
        copy._add_to_s(other)
        copy._add_to_iw(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(
            self, deep=False, x=self.x[index], y=self.y[index], s=self.s[index], iw=self.iw[index]
        )


@dataclass
class TernarySample(_BinarySampleMixin, _SubgroupSampleMixin, SampleBase[X]):
    @overload
    def add_field(self, iw: None = ...) -> Self:
        ...

    @overload
    def add_field(self, iw: Tensor) -> TernarySampleIW:
        ...

    def add_field(self, iw: Optional[Tensor] = None) -> Union[Self, TernarySampleIW]:
        if iw is not None:
            return TernarySampleIW(x=self.x, s=self.s, y=self.y, iw=iw)
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.y, self.s)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.y) > 1)
        copy._add_to_y(other)
        copy._add_to_s(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index], y=self.y[index], s=self.s[index])


@dataclass
class BinarySampleIW(_IwMixin, _BinarySampleMixin, SampleBase[X]):
    @overload
    def add_field(self, s: None = ...) -> Self:
        ...

    @overload
    def add_field(self, s: Tensor) -> TernarySampleIW:
        ...

    def add_field(self, s: Optional[Tensor] = None) -> Union[Self, TernarySampleIW]:
        if s is not None:
            return TernarySampleIW(x=self.x, s=s, y=self.y, iw=self.iw)
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.y, self.iw)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.y) > 1)
        copy._add_to_y(other)
        copy._add_to_iw(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index], y=self.y[index], iw=self.iw[index])


@dataclass
class BinarySample(_BinarySampleMixin, SampleBase[X]):
    @overload
    def add_field(self, *, s: None = ..., iw: None = ...) -> Self:
        ...

    @overload
    def add_field(self, *, s: None = ..., iw: Tensor) -> BinarySampleIW:
        ...

    @overload
    def add_field(self, *, s: Tensor, iw: None = ...) -> TernarySample:
        ...

    @overload
    def add_field(self, *, s: Tensor, iw: Tensor) -> TernarySampleIW:
        ...

    def add_field(
        self, *, s: Optional[Tensor] = None, iw: Optional[Tensor] = None
    ) -> Union[Self, BinarySampleIW, TernarySample, TernarySampleIW]:
        if s is not None:
            if iw is not None:
                return TernarySampleIW(x=self.x, s=s, y=self.y, iw=iw)
            return TernarySample(x=self.x, s=s, y=self.y)
        if iw is not None:
            return BinarySampleIW(x=self.x, y=self.y, iw=iw)
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.y)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.y) > 1)
        copy._add_to_y(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index], y=self.y[index])


@dataclass
class SubgroupSampleIW(SampleBase[X], _SubgroupSampleMixin, _IwMixin):
    @overload
    def add_field(self, y: None = ...) -> Self:
        ...

    @overload
    def add_field(self, y: Tensor) -> TernarySampleIW:
        ...

    def add_field(self, y: Optional[Tensor] = None) -> Union[Self, TernarySampleIW]:
        if y is not None:
            return TernarySampleIW(x=self.x, s=self.s, y=y, iw=self.iw)
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.s, self.iw)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.s) > 1)
        copy._add_to_s(other)
        copy._add_to_iw(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index], s=self.s[index], iw=self.iw[index])


@dataclass
class SubgroupSample(_SubgroupSampleMixin, SampleBase[X]):
    @overload
    def add_field(self, *, y: None = ..., iw: None = ...) -> Self:
        ...

    @overload
    def add_field(self, *, y: None = ..., iw: Tensor) -> SubgroupSampleIW:
        ...

    @overload
    def add_field(self, *, y: Tensor, iw: None = ...) -> TernarySample:
        ...

    @overload
    def add_field(self, *, y: Tensor, iw: Tensor) -> TernarySampleIW:
        ...

    def add_field(
        self, *, y: Optional[Tensor] = None, iw: Optional[Tensor] = None
    ) -> Union[Self, SubgroupSampleIW, TernarySample, TernarySampleIW]:
        if y is not None:
            if iw is not None:
                return TernarySampleIW(x=self.x, s=self.s, y=y, iw=iw)
            return TernarySample(x=self.x, s=self.s, y=y)
        if iw is not None:
            return SubgroupSampleIW(x=self.x, s=self.s, iw=iw)
        return self

    @override
    def __iter__(self) -> Iterator[Union[X, Tensor]]:
        yield from (self.x, self.s)

    @override
    def __add__(self, other: Self) -> Self:
        copy = self._get_copy(other, is_batched=len(self.s) > 1)
        copy._add_to_s(other)
        return copy

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index], s=self.s[index])


@dataclass
class NamedSample(SampleBase[X]):
    @overload
    def add_field(self, *, y: None = ..., s: None = ..., iw: None = ...) -> Self:
        ...

    @overload
    def add_field(self, *, y: Tensor, s: None = ..., iw: None = ...) -> BinarySample:
        ...

    @overload
    def add_field(self, *, y: Tensor, s: None = ..., iw: Tensor) -> BinarySampleIW:
        ...

    @overload
    def add_field(self, *, y: Tensor, s: Tensor, iw: None = ...) -> TernarySample:
        ...

    @overload
    def add_field(self, *, y: Tensor, s: Tensor, iw: Tensor) -> TernarySampleIW:
        ...

    def add_field(
        self, *, y: Optional[Tensor] = None, s: Optional[Tensor] = None, iw: Optional[Tensor] = None
    ) -> Union[Self, BinarySample, BinarySampleIW, TernarySample, TernarySampleIW]:
        if y is not None:
            if s is not None:
                if iw is not None:
                    return TernarySampleIW(x=self.x, s=s, y=y, iw=iw)
                return TernarySample(x=self.x, s=s, y=y)
            if iw is not None:
                return BinarySampleIW(x=self.x, y=y, iw=iw)
            return BinarySample(x=self.x, y=y)
        return self

    @override
    def __iter__(self) -> Iterator[X]:
        yield self.x

    @override
    def __getitem__(self, index: IndexType) -> Self:
        assert isinstance(self.x, (Tensor, np.ndarray)), "x is not indexable"
        return gcopy(self, deep=False, x=self.x[index])


def shallow_astuple(dataclass: object) -> Tuple[Any, ...]:
    """dataclasses.astuple() but without the deep-copying/recursion." """
    if not is_dataclass(dataclass):
        raise TypeError("shallow_astuple() should be called on dataclass instances")
    return tuple(getattr(dataclass, field.name) for field in fields(dataclass))


def shallow_asdict(dataclass: object) -> Dict[str, Any]:
    """dataclasses.asdict() but without the deep-copying/recursion." """
    if not is_dataclass(dataclass):
        raise TypeError("shallow_asdict() should be called on dataclass instances")
    return {field.name: getattr(dataclass, field.name) for field in fields(dataclass)}


@dataclass
class ImageSize(Sequence):
    c: int
    h: int
    w: int

    def __mul__(self, other: Union[Self, float]) -> Self:
        copy = gcopy(self, deep=False)
        if isinstance(other, (float, int)):
            copy.c = round(copy.c * other)
            copy.h = round(copy.h * other)
            copy.w = round(copy.w * other)
        else:
            copy.c *= other.c
            copy.h *= other.h
            copy.w *= other.w
        return copy

    def __iter__(self) -> Iterator[int]:
        yield from (self.c, self.h, self.w)

    @overload
    def __getitem__(self, index: int) -> int:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[int]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[int, Sequence[int]]:
        return (self.c, self.h, self.w)[index]

    def __len__(self) -> int:
        return 3

    @property
    def numel(self) -> int:
        return sum(iter(self))


@dataclass(kw_only=True)
class MeanStd:
    mean: Union[Tuple[float, ...], List[float]]
    std: Union[Tuple[float, ...], List[float]]

    def __iter__(self) -> Iterator[Union[Tuple[float, ...], List[float]]]:
        yield from (self.mean, self.std)

    def __imul__(self, value: float) -> Self:
        self.mean = [value * elem for elem in self.mean]
        self.std = [value * elem for elem in self.std]
        return self

    def __mul__(self, value: float) -> Self:
        copy = gcopy(self, deep=True)
        copy *= value
        return copy

    def __idiv__(self, value: float) -> Self:
        self *= 1 / value
        return self

    def __div__(self, value: float) -> Self:
        copy = gcopy(self, deep=True)
        copy *= 1 / value
        return copy


R_co = TypeVar("R_co", covariant=True)


@runtime_checkable
class Dataset(Protocol[R_co]):
    def __getitem__(self, index: int) -> R_co:
        ...


@runtime_checkable
class SizedDataset(Dataset[R_co], Sized, Protocol):
    @override
    def __getitem__(self, index: int) -> R_co:
        ...

    @override
    def __len__(self) -> Optional[int]:  # type: ignore
        ...


X2 = TypeVar("X2", bound=UnloadedData)
Y = TypeVar("Y", Tensor, None)
S = TypeVar("S", Tensor, None)


@runtime_checkable
class PseudoCdtDataset(Protocol[R_co, X2, Y, S]):
    x: X2
    y: Y
    s: S

    def __getitem__(self, index: int) -> R_co:
        ...

    def __len__(self) -> int:
        ...


D = TypeVar("D", bound=Union[Dataset, Tensor, List[int]], covariant=True)


@runtime_checkable
class DatasetWrapper(SizedDataset[R_co], Protocol):
    dataset: Dataset

    @override
    def __getitem__(self, index: int) -> R_co:
        ...

    @override
    def __len__(self) -> Optional[int]:
        if isinstance(self.dataset, SizedDataset):
            return len(self.dataset)  # type: ignore
        return None


@dataclass(kw_only=True)
class TrainTestSplit(Generic[D]):
    train: D
    test: D

    def __iter__(self) -> Iterator[D]:
        yield from (self.train, self.test)


@dataclass(kw_only=True)
class TrainValTestSplit(TrainTestSplit[D]):
    val: D

    def __iter__(self) -> Iterator[D]:
        yield from (self.train, self.val, self.test)



from dataclasses import dataclass

from ethicml.data import Dataset, Health
from ethicml.data import HealthSplits as HealthSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["HealthDataModule", "HealthSens"]


@dataclass(kw_only=True)
class HealthDataModule(EthicMlDataModule):
    """Data Module for the Heritage Health Dataset."""

    sens_feat: HealthSens = HealthSens.SEX
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Health(
            split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s
        )



from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional, Union
from typing_extensions import TypeAlias

import pandas as pd
import torch
from torch import Tensor

from conduit.data.datasets.utils import UrlFileInfo, download_from_url
from conduit.data.datasets.vision import CdtVisionDataset, ImageTform
from conduit.data.structures import TernarySample

__all__ = [
    "Waterbirds",
    "WaterbirdsSplit",
    "SampleType",
]


class WaterbirdsSplit(Enum):
    TRAIN = 0
    VAL = 1
    TEST = 2


SampleType: TypeAlias = TernarySample


class Waterbirds(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """The Waterbirds dataset introduced in `GDRO`_.

    The dataset was constructed by cropping out birds from the Caltech-UCSD Birds-200-2011 (CUB) dataset
    and transferring them onto backgrounds from the Places dataset.

    The train-test split is the same as the one used for the CUB dataset, with 20% of the training data chosen to
    serve as a validation set. For the validation and test sets, landbirds and waterbirds are equally
    distributed across land and water backgrounds (i.e., there are the same number of landbirds on land
    vs. water backgrounds, and separately, the same number of waterbirds on land vs. water backgrounds).
    This allows us to more accurately measure the performance of the rare groups, and it is particularly
    important for the Waterbirds dataset because of its relatively small size; otherwise, the smaller groups
    (waterbirds on land and landbirds on water) would have too few samples to accurately estimate performance on.

    .. _GDRO:
        Generalization<https://arxiv.org/abs/1911.08731>`__
    """

    SampleType: TypeAlias = TernarySample
    Split: TypeAlias = WaterbirdsSplit

    _BASE_DIR_NAME: ClassVar[str] = "Waterbirds"
    _ORIGINAL_METADATA_FILENAME: ClassVar[str] = "metadata.csv"
    # Path to the 'fixed' metadata featured in 'MaskTune' and downloaded from the link provided in
    # the official repo of said paper:
    # 'https://drive.google.com/file/d/1xPNYQskEXuPhuqT5Hj4hXPeJa9jh7liL/view'.
    # To quote the paper:
    # "We discovered and fixed two problems with the Waterbirds dataset: a) Because the background
    # images in the Places dataset (Zhou et al., 2017) may already contain bird images, multiple
    # birds may appear in an image after overlaying the segmented bird images from the Caltech-UCSD
    # Birds- 200-2011 (CUB) dataset (Wah et al., 2011). For example, the label of an image may be
    # “landbird”, but the image contains both land and water birds. Such images were removed from
    # the dataset. b) Because the names of the species are similar, some land birds have been
    # mislabeled as waterbirds"
    _FIXED_METADATA_FILEPATH: ClassVar[Path] = Path(__file__).parent / "waterbirds_fixed.csv.zip"

    _FILE_INFO: ClassVar[UrlFileInfo] = UrlFileInfo(
        name="Waterbirds.tar.gz",
        url="https://worksheets.codalab.org/rest/bundles/0x505056d5cdea4e4eaa0e242cbfe2daa4/contents/blob/",
        md5=None,
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        split: Optional[Union[WaterbirdsSplit, str]] = None,
        fixed: bool = False,
    ) -> None:
        self.split = WaterbirdsSplit[split.upper()] if isinstance(split, str) else split
        self.root = Path(root)
        self._base_dir = self.root / self._BASE_DIR_NAME
        self.fixed = fixed
        self.download = download
        if self.download:
            download_from_url(
                file_info=self._FILE_INFO,
                root=self.root,
                logger=self.logger,
                remove_finished=True,
            )
        elif not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )

        metadata_fp = (
            self._FIXED_METADATA_FILEPATH
            if self.fixed
            else self._base_dir / self._ORIGINAL_METADATA_FILENAME
        )
        # Read in metadata
        # Note: metadata is one-indexed.
        self.metadata = pd.read_csv(metadata_fp)
        # Use an official split of the data, if specified, else just use all
        # of the data
        if self.split is not None:
            split_indices = self.metadata["split"] == self.split.value
            self.metadata = self.metadata.loc[split_indices]

        # Extract filenames
        x = self.metadata['img_filename'].to_numpy()
        # Extract class (land- vs. water-bird) labels
        y = torch.as_tensor(self.metadata["y"].to_numpy(), dtype=torch.long)
        # Extract place (land vs. water) labels
        s = torch.as_tensor(self.metadata["place"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)

    def _check_unzipped(self) -> bool:
        return (self._base_dir).is_dir()



from typing import Optional, Tuple

import torch

from conduit.data.datasets.base import CdtDataset

__all__ = ["DummyDataset"]


class DummyDataset(CdtDataset):
    """Generate a dummy dataset.

    :param shapes: list of shapes
    :param num_samples: how many samples to use in this dataset
    :param seed: random seed
    """

    def __init__(
        self, *shapes: Tuple[int, ...], num_samples: int = 10000, seed: Optional[int] = None
    ) -> None:
        self.shapes = shapes
        self.num_samples = num_samples

        self.generator = (
            torch.default_generator if seed is None else torch.Generator().manual_seed(seed)
        )

        x = torch.randn(num_samples, *shapes[0], generator=self.generator)
        s = (torch.rand(num_samples, *shapes[1], generator=self.generator) > 0.5).float().squeeze()
        y = (torch.rand(num_samples, *shapes[1], generator=self.generator) > 0.5).float().squeeze()

        super().__init__(x=x, s=s, y=y)



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from conduit.data.datasets.vision.camelyon17 import Camelyon17Attr
from conduit.data.datasets.vision.camelyon17 import Camelyon17Split
from conduit.data.datasets.vision.camelyon17 import Camelyon17SplitScheme
from conduit.data.datasets.vision.celeba import CelebASplit
from conduit.data.datasets.vision.celeba import CelebAttr
from conduit.data.datasets.vision.isic import IsicAttr
from conduit.data.datasets.vision.nico import NicoSuperclass
from conduit.data.datasets.vision.ssrp import SSRPSplit
from conduit.data.datasets.vision.waterbirds.waterbirds import WaterbirdsSplit
from omegaconf import MISSING
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Union


@dataclass
class CelebAConf:
    _target_: str = "conduit.data.datasets.vision.CelebA"
    root: Union[Path, str] = MISSING
    download: bool = True
    superclass: Union[CelebAttr, str] = CelebAttr.SMILING
    subclass: Union[CelebAttr, str] = CelebAttr.MALE
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    split: Optional[Union[CelebASplit, str]] = None


@dataclass
class ColoredMNISTConf:
    _target_: str = "conduit.data.datasets.vision.ColoredMNIST"
    root: Union[Path, str] = MISSING
    download: bool = True
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    label_map: Any = None  # Optional[Dict[int, int]]
    colors: Optional[List[int]] = None
    num_colors: int = 10
    scale: float = 0.2
    correlation: Optional[float] = None
    binarize: bool = False
    greyscale: bool = False
    background: bool = False
    black: bool = True
    split: Any = None  # Optional[Union[ColoredMNISTSplit, List[int], str]]
    seed: Optional[int] = 42


@dataclass
class ISICConf:
    _target_: str = "conduit.data.datasets.vision.ISIC"
    root: Union[Path, str] = MISSING
    download: bool = True
    max_samples: int = 25000
    target_attr: IsicAttr = IsicAttr.MALIGNANT
    context_attr: IsicAttr = IsicAttr.HISTO
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]


@dataclass
class NICOConf:
    _target_: str = "conduit.data.datasets.vision.NICO"
    root: Union[Path, str] = MISSING
    download: bool = True
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    superclass: Optional[Union[NicoSuperclass, str]] = NicoSuperclass.ANIMALS


@dataclass
class SSRPConf:
    _target_: str = "conduit.data.datasets.vision.SSRP"
    root: Union[Path, str] = MISSING
    split: Union[SSRPSplit, str] = SSRPSplit.PRETRAIN
    download: bool = True
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]


@dataclass
class WaterbirdsConf:
    _target_: str = "conduit.data.datasets.vision.Waterbirds"
    root: Union[Path, str] = MISSING
    download: bool = True
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    split: Optional[Union[WaterbirdsSplit, str]] = None
    fixed: bool = False


@dataclass
class Camelyon17Conf:
    _target_: str = "conduit.data.datasets.vision.Camelyon17"
    root: Union[Path, str] = MISSING
    download: bool = True
    transform: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    split: Optional[Union[Camelyon17Split, str]] = None
    split_scheme: Union[Camelyon17SplitScheme, str] = Camelyon17SplitScheme.OFFICIAL
    superclass: Union[Camelyon17Attr, str] = Camelyon17Attr.TUMOR
    subclass: Union[Camelyon17Attr, str] = Camelyon17Attr.CENTER



import random
from typing import Sequence, Tuple, Union

from PIL import Image, ImageFilter, ImageOps
import torch
from torch import Tensor
import torchvision.transforms as T  # type: ignore
import torchvision.transforms.functional as TF  # type: ignore

from conduit.types import NDArrayR

__all__ = [
    "Denormalize",
    "RandomGaussianBlur",
    "RandomSolarize",
    "denormalize",
]


def _invert_norm_values(
    mean: Union[float, Tensor, Sequence[float], NDArrayR],
    std: Union[float, Tensor, Sequence[float], NDArrayR],
) -> Tuple[Tensor, Tensor]:
    mean = torch.as_tensor(mean)
    std = torch.as_tensor(std)
    std_inv = std.reciprocal().clip(torch.finfo(std.dtype).eps)
    mean_inv = -mean * std_inv
    return mean_inv, std_inv


def denormalize(
    tensor: Tensor,
    mean: Union[float, Tensor, Sequence[float], NDArrayR],
    std: Union[float, Tensor, Sequence[float], NDArrayR],
    inplace: bool = False,
) -> Tensor:
    """Denormalize a float tensor image with mean and standard deviation.

    Given mean: ``(mean[1],...,mean[n])`` and std: ``(std[1],..,std[n])`` for ``n``
    channels, this transform will dennormalize each channel of the input
    ``torch.*Tensor`` i.e.,
    ``output[channel] = (input[channel] * std[channel]) + mean[channel])``

    .. note::
        This transform acts out of place, i.e., it does not mutate the input tensor.
    """
    mean_inv, std_inv = _invert_norm_values(mean=mean, std=std)
    return TF.normalize(
        tensor=tensor, mean=mean_inv.tolist(), std=std_inv.tolist(), inplace=inplace
    )


class Denormalize(T.Normalize):
    """Denormalize a float tensor image with mean and standard deviation.

    Given mean: ``(mean[1],...,mean[n])`` and std: ``(std[1],..,std[n])`` for ``n``
    channels, this transform will dennormalize each channel of the input
    ``torch.*Tensor`` i.e.,
    ``output[channel] = (input[channel] * std[channel]) + mean[channel])``

    .. note::
        This transform acts out of place, i.e., it does not mutate the input tensor.
    """

    def __init__(
        self,
        mean: Union[float, Tensor, Sequence[float], NDArrayR],
        std: Union[float, Tensor, Sequence[float], NDArrayR],
        inplace: bool = False,
    ) -> None:
        mean_inv, std_inv = _invert_norm_values(mean=mean, std=std)
        super().__init__(mean=mean_inv, std=std_inv, inplace=inplace)


class RandomGaussianBlur:
    """
    Apply Gaussian Blur to the PIL image with some probability.
    """

    def __init__(self, p: float = 0.5, *, radius_min: float = 0.1, radius_max: float = 2.0) -> None:
        self.prob = p
        self.radius_min = radius_min
        self.radius_max = radius_max

    def __call__(self, img: Image.Image) -> Image.Image:
        do_it = random.random() <= self.prob
        if not do_it:
            return img

        return img.filter(
            ImageFilter.GaussianBlur(radius=random.uniform(self.radius_min, self.radius_max))
        )


class RandomSolarize:
    """
    Apply Solarization to a PIL image with some probability.
    """

    def __init__(self, p: float) -> None:
        self.p = p

    def __call__(self, img: Image.Image) -> Image.Image:
        return ImageOps.solarize(img) if random.random() < self.p else img



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from conduit.data.datasets.vision.camelyon17 import Camelyon17Attr
from conduit.data.datasets.vision.camelyon17 import Camelyon17SplitScheme
from conduit.data.datasets.vision.celeba import CelebAttr
from conduit.data.datasets.vision.nico import NicoSuperclass
from omegaconf import MISSING
from pathlib import Path
from ranzen.torch.data import TrainingMode
from typing import Any
from typing import List
from typing import Optional
from typing import Union


@dataclass
class CelebADataModuleConf:
    _target_: str = "conduit.data.datamodules.vision.CelebADataModule"
    root: Union[Path, str] = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    test_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    image_size: int = 224
    superclass: CelebAttr = CelebAttr.SMILING
    subclass: CelebAttr = CelebAttr.MALE
    use_predefined_splits: bool = False


@dataclass
class ColoredMNISTDataModuleConf:
    _target_: str = "conduit.data.datamodules.vision.ColoredMNISTDataModule"
    root: Union[Path, str] = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    test_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    image_size: int = 32
    use_predefined_splits: bool = False
    label_map: Any = None  # Optional[Dict[int, int]]
    colors: Optional[List[int]] = None
    num_colors: int = 10
    scale: float = 0.2
    correlation: float = 1.0
    binarize: bool = False
    greyscale: bool = False
    background: bool = False
    black: bool = True


@dataclass
class NICODataModuleConf:
    _target_: str = "conduit.data.datamodules.vision.NICODataModule"
    root: Union[Path, str] = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    test_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    image_size: int = 224
    class_train_props: Optional[dict] = None
    superclass: NicoSuperclass = NicoSuperclass.ANIMALS


@dataclass
class WaterbirdsDataModuleConf:
    _target_: str = "conduit.data.datamodules.vision.WaterbirdsDataModule"
    root: Union[Path, str] = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    test_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    image_size: int = 224
    use_predefined_splits: bool = False


@dataclass
class Camelyon17DataModuleConf:
    _target_: str = "conduit.data.datamodules.vision.Camelyon17DataModule"
    root: Union[Path, str] = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    test_tf: Any = None  # Optional[Union[BasicTransform, Callable[[Image], Any], Compose]]
    image_size: int = 96
    superclass: Camelyon17Attr = Camelyon17Attr.TUMOR
    subclass: Camelyon17Attr = Camelyon17Attr.CENTER
    use_predefined_splits: bool = False
    split_scheme: Camelyon17SplitScheme = Camelyon17SplitScheme.OFFICIAL



from enum import Enum
from typing import Any, Callable, Dict, Optional, Union, cast
from typing_extensions import TypeAlias, override

import pytorch_lightning as pl
from pytorch_lightning.callbacks.progress.rich_progress import (
    BatchesProcessedColumn,
    CustomBarColumn,
    CustomTimeColumn,
    ProcessingSpeedColumn,
    RichProgressBar,
    RichProgressBarTheme,
)
from pytorch_lightning.utilities.types import STEP_OUTPUT
from rich.console import RenderableType
from rich.progress import Task, TaskID, TextColumn
from rich.table import Column
from rich.text import Text

__all__ = ["CdtProgressBar", "ProgressBarTheme"]


class _FixedLengthProcessionSpeed(ProcessingSpeedColumn):
    """Renders processing speed for the progress bar with fixes length"""

    def __init__(self, style: Union[str, Any]) -> None:
        super().__init__(style)
        self.max_length = len("0.00")

    def render(self, task: Task) -> RenderableType:
        task_speed = f"{task.speed:>.2f}" if task.speed is not None else "0.00"
        self.max_length = max(len(task_speed), self.max_length)
        task_speed = " " * (self.max_length - len(task_speed)) + task_speed
        return Text(f"{task_speed}it/s", style=self.style, justify="center")


def _quaterion_theme() -> RichProgressBarTheme:
    return RichProgressBarTheme(
        description="white",
        progress_bar="#4881AD",
        progress_bar_finished="#67C87A",
        progress_bar_pulse="#67C87A",
        batch_progress="white",
        time="grey54",
        processing_speed="grey70",
        metrics="white",
    )


def _cyberpunk_theme() -> RichProgressBarTheme:
    return RichProgressBarTheme(
        description="#FF00CF",
        progress_bar="#001eff",
        progress_bar_finished="#00ff9f",
        progress_bar_pulse="#001eff",
        batch_progress="#FF00CF",
        time="#00b8ff",
        processing_speed="#FF46D6",
        metrics="#00ff9f",
    )


def _google_theme() -> RichProgressBarTheme:
    return RichProgressBarTheme(
        description="#008744",
        progress_bar="#0057e7",
        progress_bar_finished="#d62d20",
        progress_bar_pulse="#d62d20",
        batch_progress="#ffa700",
        time="white",
        processing_speed="#0057e7",
        metrics="#d62d20",
    )


class ProgressBarTheme(Enum):
    QUATERION = (_quaterion_theme,)
    CYBERPUNK = (_cyberpunk_theme,)
    GOOGLE = (_google_theme,)

    def __init__(self, load: Callable[..., RichProgressBarTheme]) -> None:
        self.load = load


class CdtProgressBar(RichProgressBar):
    """Custom Lightning progress bar supporting both epoch- and iteration-based training."""

    Theme: TypeAlias = ProgressBarTheme

    def __init__(
        self,
        refresh_rate: int = 1,
        *,
        leave: bool = False,
        theme: Union[RichProgressBarTheme, Theme] = Theme.CYBERPUNK,
        console_kwargs: Optional[Dict[str, Any]] = None,
        predict_description: str = "Predicting",
    ) -> None:
        if isinstance(theme, ProgressBarTheme):
            theme = cast(RichProgressBarTheme, theme.load())

        super().__init__(
            refresh_rate=refresh_rate,
            leave=leave,
            theme=theme,
            console_kwargs=console_kwargs,
        )
        self.predict_progress_bar_id = None
        self._predict_description = predict_description

    @override
    def configure_columns(self, *args: Any, **kwargs: Any) -> list:
        return [
            TextColumn(
                "[progress.description]{task.description}",
                table_column=Column(
                    no_wrap=True,
                    min_width=9,  # prevents blinking during validation, length of `Validation `
                ),
            ),
            CustomBarColumn(
                complete_style=self.theme.progress_bar,
                finished_style=self.theme.progress_bar_finished,
                pulse_style=self.theme.progress_bar_pulse,
            ),
            BatchesProcessedColumn(style=self.theme.batch_progress),
            CustomTimeColumn(style=self.theme.time),
            _FixedLengthProcessionSpeed(style=self.theme.processing_speed),
        ]

    @staticmethod
    def _is_iteration_based(trainer: pl.Trainer) -> bool:
        return trainer.num_training_batches == float("inf")

    @property
    @override
    def total_train_batches(self) -> Union[int, float]:
        """
        The total number of training batches, which may change from epoch to epoch.

        Use this to set the total number of iterations in the progress bar. Can return ``inf`` if
        the training dataloader is of infinite size.
        """
        if self.trainer.max_steps >= 0:
            return self.trainer.max_steps
        return self.trainer.num_training_batches

    @override
    def _add_task(  # type: ignore
        self, total_batches: float, description: str, visible: bool = True
    ) -> Optional[TaskID]:
        if self.progress is not None:
            return self.progress.add_task(
                f"[{self.theme.description}]{description}", total=total_batches, visible=visible
            )

    @override
    def _update(
        self, progress_bar_id: Optional[TaskID], current: int, visible: bool = True
    ) -> None:
        if self.progress is not None and self.is_enabled:
            assert progress_bar_id is not None
            total = self.progress.tasks[progress_bar_id].total
            if not self._should_update(current, total):  # type: ignore
                return

            leftover = current % self.refresh_rate
            advance = leftover if (current == total and leftover != 0) else self.refresh_rate
            self.progress.update(progress_bar_id, advance=advance, visible=visible)
            self.refresh()

    @override
    def _get_train_description(self, current_epoch: Optional[int]) -> str:
        train_description = f"Training"
        if current_epoch is not None:
            train_description += f" (Epoch: {current_epoch})"
        if len(self.validation_description) > len(train_description):
            # Padding is required to avoid flickering due of uneven lengths of "Epoch X"
            # and "Validation" Bar description
            num_digits = len(str(current_epoch))
            required_padding = (
                len(self.validation_description) - len(train_description) + 1
            ) - num_digits
            for _ in range(required_padding):
                train_description += " "
        return train_description

    @property
    def _progress_bar_id(self) -> Optional[TaskID]:
        return getattr(self, "train_progress_bar_id", getattr(self, "main_progress_bar_id", None))

    @_progress_bar_id.setter
    def _progress_bar_id(self, value: Optional[TaskID]) -> None:
        if hasattr(self, "train_progress_bar_id"):
            self.train_progress_bar_id = value
        if hasattr(self, "main_progress_bar_id"):
            self.main_progress_bar_id = value

    @override
    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        total_train_batches = self.total_train_batches
        total_val_batches = self.total_val_batches
        if total_train_batches != float("inf"):
            # val can be checked multiple times per epoch
            val_checks_per_epoch = total_train_batches // trainer.val_check_batch
            total_val_batches = total_val_batches * val_checks_per_epoch

        iteration_based = self._is_iteration_based(trainer)
        current_epoch = None if iteration_based else trainer.current_epoch
        train_description = self._get_train_description(current_epoch)
        if self._progress_bar_id is not None and self._leave:
            self._stop_progress()
            self._init_progress(trainer)
        if self._progress_bar_id is None:
            self._progress_bar_id = self._add_task(total_train_batches, train_description)
        elif (not iteration_based) and (self.progress is not None):
            self.progress.reset(
                TaskID(self._progress_bar_id),
                total=total_train_batches,
                description=train_description,
                visible=True,
            )
        self.refresh()

    @override
    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if trainer.sanity_checking:
            self._update(self.val_sanity_progress_bar_id, batch_idx + 1)
        elif self.val_progress_bar_id is not None:
            # check to see if we should update the main training progress bar
            self._update(self.val_progress_bar_id, batch_idx + 1)
        self.refresh()

    @property
    def predict_description(self) -> str:
        return self._predict_description

    @predict_description.setter
    def predict_description(self, value: str) -> None:
        self._predict_description = value

    @override
    def on_predict_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if self.has_dataloader_changed(dataloader_idx):
            if (self.predict_progress_bar_id is not None) and (self.progress is not None):
                self.progress.update(TaskID(self.predict_progress_bar_id), advance=0, visible=False)
            self.predict_progress_bar_id = self._add_task(
                self.total_predict_batches_current_dataloader, self.predict_description
            )
            self.refresh()



"""Dummy data-module."""

from dataclasses import dataclass
from pathlib import Path
from typing import Union
from typing_extensions import override

from conduit.data import TrainValTestSplit
from conduit.data.datamodules.vision import CdtVisionDataModule
from conduit.data.datasets.vision.dummy import DummyVisionDataset, SampleType


@dataclass(kw_only=True)
class DummyVisionDataModule(CdtVisionDataModule[SampleType]):
    num_samples: int = 1_000
    seed: int = 8
    root: Union[str, Path] = ""
    height: int = 32
    width: int = 32
    channels: int = 3
    batch_size: int = 32
    s_card: int | None = 2
    y_card: int | None = 2

    @override
    def _get_image_splits(self) -> TrainValTestSplit[DummyVisionDataset]:
        # Split the data randomly according to val- and test-prop
        data = DummyVisionDataset(
            channels=self.channels,
            height=self.height,
            width=self.width,
            num_samples=self.num_samples,
            s_card=self.s_card,
            y_card=self.y_card,
        )
        val_data, test_data, train_data = data.random_split(props=(self.val_prop, self.test_prop))
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



import logging
import sys
from typing import Optional

__all__ = ["init_logger"]


def init_logger(name: Optional[str] = None, level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(level)
    return logger



"""CelebA Dataset."""

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional, Union
from typing_extensions import TypeAlias

import numpy as np
from numpy import typing as npt
import pandas as pd
import torch
from torch import Tensor

from conduit.data.datasets import GdriveFileInfo, download_from_gdrive
from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["CelebA", "CelebAttr", "CelebASplit"]


class CelebAttr(Enum):
    FIVE_O_CLOCK_SHADOW = "Five_o_Clock_Shadow"
    ARCHED_EYEBROWS = "Arched_Eyebrows"
    ATTRACTIVE = "Attractive"
    BAGS_UNDER_EYES = "Bags_Under_Eyes"
    BALD = "Bald"
    BANGS = "Bangs"
    BIG_LIPS = "Big_Lips"
    BIG_NOSE = "Big_Nose"
    BLACK_HAIR = "Black_Hair"
    BLOND_HAIR = "Blond_Hair"
    BLURRY = "Blurry"
    BROWN_HAIR = "Brown_Hair"
    BUSHY_EYEBROWS = "Bushy_Eyebrows"
    CHUBBY = "Chubby"
    DOUBLE_CHIN = "Double_Chin"
    EYEGLASSES = "Eyeglasses"
    GOATEE = "Goatee"
    GRAY_HAIR = "Gray_Hair"
    HEAVY_MAKEUP = "Heavy_Makeup"
    HIGH_CHEEKBONES = "High_Cheekbones"
    MALE = "Male"
    MOUTH_SLIGHTLY_OPEN = "Mouth_Slightly_Open"
    MUSTACHE = "Mustache"
    NARROW_EYES = "Narrow_Eyes"
    NO_BEARD = "No_Beard"
    OVAL_FACE = "Oval_Face"
    PALE_SKIN = "Pale_Skin"
    POINTY_NOSE = "Pointy_Nose"
    RECEDING_HAIRLINE = "Receding_Hairline"
    ROSY_CHEEKS = "Rosy_Cheeks"
    SIDEBURNS = "Sideburns"
    SMILING = "Smiling"
    STRAIGHT_HAIR = "Straight_Hair"
    WAVY_HAIR = "Wavy_Hair"
    WEARING_EARRINGS = "Wearing_Earrings"
    WEARING_HAT = "Wearing_Hat"
    WEARING_LIPSTICK = "Wearing_Lipstick"
    WEARING_NECKLACE = "Wearing_Necklace"
    WEARING_NECKTIE = "Wearing_Necktie"
    YOUNG = "Young"


class CelebASplit(Enum):
    TRAIN = 0
    VAL = 1
    TEST = 2


SampleType: TypeAlias = TernarySample


class CelebA(CdtVisionDataset[SampleType, Tensor, Tensor]):
    """CelebA dataset."""

    SampleType: TypeAlias = TernarySample
    Attr: TypeAlias = CelebAttr
    Split: TypeAlias = CelebASplit

    # The data is downloaded to `download_dir` / `CelebA` / `_IMAGE_DIR`.
    _IMAGE_DIR: ClassVar[str] = "img_align_celeba"
    # Google drive IDs, MD5 hashes and filenames for the CelebA files.
    _FILE_LIST: ClassVar[List[GdriveFileInfo]] = [
        GdriveFileInfo(
            name="img_align_celeba.zip",
            id="1zmsC4yvw-e089uHXj5EdP0BSZ0AlDQRR",
            md5="00d2c5bc6d35e252742224ab0c1e8fcb",
        ),
        GdriveFileInfo(
            name="list_attr_celeba.txt",
            id="1gxmFoeEPgF9sT65Wpo85AnHl3zsQ4NvS",
            md5="75e246fa4810816ffd6ee81facbd244c",
        ),
        GdriveFileInfo(
            name="list_eval_partition.txt",
            id="1ih_VMokoI774ErNWrb26lDeWlanUBpnX",
            md5="d32c9cbf5e040fd4025c592c306e6668",
        ),
    ]

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        superclass: Union[CelebAttr, str] = CelebAttr.SMILING,
        subclass: Union[CelebAttr, str] = CelebAttr.MALE,
        transform: Optional[ImageTform] = None,
        split: Optional[Union[CelebASplit, str]] = None,
    ) -> None:
        self.superclass = CelebAttr(superclass) if isinstance(superclass, str) else superclass
        self.subclass = CelebAttr(subclass) if isinstance(subclass, str) else subclass
        self.split = CelebASplit[split.upper()] if isinstance(split, str) else split

        self.root = Path(root)
        self._base_dir = self.root / self.__class__.__name__
        image_dir = self._base_dir / self._IMAGE_DIR

        if download:
            download_from_gdrive(file_info=self._FILE_LIST, root=self._base_dir, logger=self.logger)
        elif not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )

        if self.split is None:
            skiprows = None
        else:
            # splits: information about which samples belong to train, val or test
            splits = (
                pd.read_csv(
                    self._base_dir / "list_eval_partition.txt",
                    delim_whitespace=True,
                    names=["split"],
                )
                .to_numpy()
                .squeeze()
            )
            skiprows = (splits != self.split.value).nonzero()[0] + 2
        attrs = pd.read_csv(
            self._base_dir / "list_attr_celeba.txt",
            delim_whitespace=True,
            header=1,
            usecols=[self.superclass.value, self.subclass.value],  # type: ignore
            skiprows=skiprows,  # type: ignore
        )

        x: npt.NDArray[np.bytes_] = np.array(attrs.index)
        s_unmapped = torch.as_tensor(attrs[self.subclass.value].to_numpy())
        y_unmapped = torch.as_tensor(attrs[self.superclass.value].to_numpy())
        # map from {-1, 1} to {0, 1}
        s_binary = torch.div(s_unmapped + 1, 2, rounding_mode="floor")
        y_binary = torch.div(y_unmapped + 1, 2, rounding_mode="floor")

        super().__init__(x=x, y=y_binary, s=s_binary, transform=transform, image_dir=image_dir)

    def _check_unzipped(self) -> bool:
        return (self._base_dir / self._IMAGE_DIR).is_dir()



"""Ecoacoustics data-module."""

from dataclasses import dataclass
from typing import Any, List, Optional
from typing_extensions import override

from torch.utils.data import Sampler

from conduit.data.datasets.audio.ecoacoustics import Ecoacoustics, SoundscapeAttr
from conduit.data.datasets.utils import AudioTform, CdtDataLoader
from conduit.data.datasets.wrappers import AudioTransformer
from conduit.data.structures import BinarySample, TrainValTestSplit
from conduit.transforms.audio import Compose, Framing, LogMelSpectrogram

from .base import CdtAudioDataModule

__all__ = ["EcoacousticsDataModule"]


@dataclass(kw_only=True)
class EcoacousticsDataModule(CdtAudioDataModule[BinarySample]):
    """Data-module for the Ecoacoustics dataset."""

    segment_len: float = 15
    sample_rate: int = 48_000
    target_attrs: List[SoundscapeAttr]

    @staticmethod
    def _batch_converter(batch: BinarySample) -> BinarySample:
        return BinarySample(x=batch.x, y=batch.y)

    @override
    def make_dataloader(
        self,
        ds: AudioTransformer,
        *,
        batch_size: int,
        shuffle: bool = False,
        drop_last: bool = False,
        batch_sampler: Optional[Sampler[List[int]]] = None,
    ) -> CdtDataLoader[BinarySample]:
        """Make DataLoader."""
        return CdtDataLoader(
            ds,
            batch_size=batch_size if batch_sampler is None else 1,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=drop_last,
            persistent_workers=self.persist_workers,
            batch_sampler=batch_sampler,
            converter=self._batch_converter,
        )

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        Ecoacoustics(
            root=self.root,
            download=True,
            segment_len=self.segment_len,
            target_attrs=self.target_attrs,
        )

    @property
    def _default_transform(self) -> Compose:
        return Compose([LogMelSpectrogram(), Framing()])

    @property
    @override
    def _default_train_transforms(self) -> AudioTform:  # type: ignore
        return self._default_transform

    @property
    @override
    def _default_test_transforms(self) -> AudioTform:  # type: ignore
        return self._default_transform

    @override
    def _get_audio_splits(self) -> TrainValTestSplit[Ecoacoustics]:
        all_data = Ecoacoustics(
            root=self.root,
            transform=None,  # Transform is applied in `CdtAudioDataModule._setup`
            segment_len=self.segment_len,
            target_attrs=self.target_attrs,
            sample_rate=self.sample_rate,
            download=False,
        )

        val_data, test_data, train_data = all_data.random_split(
            props=(self.val_prop, self.test_prop), seed=self.seed
        )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



"""Ecoacostics dataset provided A. Eldridge et al.
    Alice Eldridge, Paola Moscoso, Patrice Guyot, & Mika Peck. (2018).
    Data for "Sounding out Ecoacoustic Metrics: Avian species richness
    is predicted by acoustic indices in temperate but not tropical
    habitats" (Final) [Data set].
    Zenodo. https://doi.org/10.5281/zenodo.1255218
"""

from enum import auto
import math
from pathlib import Path
import shutil
from typing import ClassVar, Final, List, Optional, Tuple, Union
from typing_extensions import TypeAlias, override
import zipfile

import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype
from ranzen import StrEnum
import torch
from torch import Tensor
import torchaudio  # type: ignore
from tqdm import tqdm

from conduit.data.datasets.audio.base import CdtAudioDataset
from conduit.data.datasets.utils import AudioTform, UrlFileInfo, download_from_url
from conduit.data.structures import BinarySample
from conduit.types import IndexType

__all__ = ["Ecoacoustics", "SoundscapeAttr"]


class SoundscapeAttr(StrEnum):
    HABITAT = auto()
    SITE = auto()
    TIME = auto()
    NN = "NN"
    N0 = "N0"


SampleType: TypeAlias = BinarySample


class Ecoacoustics(CdtAudioDataset[SampleType, Tensor, Tensor]):
    """Dataset for audio data collected in various geographic locations."""

    _INDICES_DIR: ClassVar[str] = "AvianID_AcousticIndices"
    _METADATA_FILENAME: ClassVar[str] = "metadata.csv"
    _PBAR_COL: ClassVar[str] = "#00FF00"

    _EC_LABELS_FILENAME: ClassVar[str] = "EC_AI.csv"
    _UK_LABELS_FILENAME: ClassVar[str] = "UK_AI.csv"

    _FILE_INFO: List[UrlFileInfo] = [
        UrlFileInfo(
            name="AvianID_AcousticIndices.zip",
            url="https://zenodo.org/record/1255218/files/AvianID_AcousticIndices.zip",
            md5="b23208eb7db3766a1d61364b75cb4def",
        ),
        UrlFileInfo(
            name="EC_BIRD.zip",
            url="https://zenodo.org/record/1255218/files/EC_BIRD.zip",
            md5="d427e904af1565dbbfe76b05f24c258a",
        ),
        UrlFileInfo(
            name="UK_BIRD.zip",
            url="https://zenodo.org/record/1255218/files/UK_BIRD.zip",
            md5="e1e58b224bb8fb448d1858b9c9ee0d8c",
        ),
    ]

    num_frames_in_segment: int
    _MAX_AUDIO_LEN: Final[int] = 60

    def __init__(
        self,
        root: str,
        *,
        target_attrs: List[SoundscapeAttr],
        transform: Optional[AudioTform] = None,
        download: bool = True,
        segment_len: float = 15,
        sample_rate: int = 48_000,  # This is the value that is present in the dataset
    ) -> None:
        self.root = Path(root).expanduser()
        self.download = download
        self.base_dir = self.root / self.__class__.__name__
        self.labels_dir = self.base_dir / self._INDICES_DIR / self._INDICES_DIR
        self.sample_rate = sample_rate  # set prior to segment length
        self.segment_len = segment_len
        self._metadata_path = self.base_dir / self._METADATA_FILENAME
        self.ec_labels_path = self.labels_dir / self._EC_LABELS_FILENAME
        self.uk_labels_path = self.labels_dir / self._UK_LABELS_FILENAME
        # map provided attributes (as string or enum) to permitted attributes (as strings)
        self.target_attrs = [str(SoundscapeAttr(elem)) for elem in target_attrs]

        if self.download:
            self._download_files()
        self._check_files()

        # Extract labels from indices files.
        if not self._metadata_path.exists():
            self._extract_metadata()

        self.metadata = pd.read_csv(self.base_dir / self._METADATA_FILENAME)

        x = self.metadata["filePath"].to_numpy()
        y = torch.as_tensor(
            self._label_encode(self.metadata[self.target_attrs], inplace=True).to_numpy()
        )

        super().__init__(x=x, y=y, transform=transform, audio_dir=self.base_dir)

    @property
    def segment_len(self) -> float:
        return self._segment_len

    @segment_len.setter
    def segment_len(self, value: float) -> None:
        if value <= 0:
            raise ValueError("Segment length must be positive.")
        value = min(value, self._MAX_AUDIO_LEN)
        self._segment_len = value
        self.num_frames_in_segment = int(self.segment_len * self.sample_rate)

    def _check_files(self) -> None:
        """Check necessary files are present and unzipped."""

        if not self.labels_dir.exists():
            raise FileNotFoundError(
                f"Indices file not found at location {self.labels_dir.resolve()}. "
                "Have you downloaded it?"
            )
        if zipfile.is_zipfile(self.labels_dir):
            raise RuntimeError("Indices file not unzipped.")

        for dir_ in ["UK_BIRD", "EC_BIRD"]:
            path = self.base_dir / dir_
            if not path.exists():
                raise RuntimeError(
                    f"Data not found at location {self.base_dir.resolve()}. Have you downloaded it?"
                )
            if zipfile.is_zipfile(dir_):
                raise RuntimeError(f"{dir_} file not unzipped.")

    @staticmethod
    def _label_encode(
        data: Union[pd.DataFrame, pd.Series], inplace: bool = True
    ) -> Union[pd.DataFrame, pd.Series]:
        """Label encode the extracted concept/context/superclass information."""
        data = data.copy(deep=not inplace)
        if isinstance(data, pd.Series):
            if is_object_dtype(data) or isinstance(data.dtype, pd.CategoricalDtype):
                data.update(data.factorize()[0])  # type: ignore
                data = data.astype(np.int64)
        else:
            for col in data.columns:
                # Add a new column containing the label-encoded data
                if is_object_dtype(data[col]) or isinstance(data.dtype, pd.CategoricalDtype):
                    data[col] = data[col].factorize()[0]
        return data

    def _download_files(self) -> None:
        """Download all files necessary for dataset to work."""
        # Create necessary directories if they don't already exist.
        self.base_dir.mkdir(parents=True, exist_ok=True)

        for finfo in self._FILE_INFO:
            download_from_url(
                file_info=finfo, root=self.base_dir, logger=self.logger, remove_finished=True
            )
        if (macosx_dir := self.base_dir / "__MACOSX").exists():
            shutil.rmtree(macosx_dir)

    def _extract_metadata(self) -> None:
        """Extract information such as labels from relevant csv files, combining them along with
        information on processed files to produce a master file."""

        self.logger.info("Extracting metadata.")
        # Process the metadata for samples from Ecuador.
        ec_labels = pd.read_csv(self.ec_labels_path, encoding="ISO-8859-1")
        ec_labels["filePath"] = "EC_BIRD/" + ec_labels["fileName"]
        # Process the metadata for samples from the UK.
        uk_labels = pd.read_csv(self.uk_labels_path, encoding="ISO-8859-1")
        # Some waveforms use full names for location i.e. BALMER and KNEPP, change indices file to do the same.
        uk_labels.replace(regex={"BA-": "BALMER-", "KN-": "KNEPP-"}, inplace=True)
        uk_labels["filePath"] = "UK_BIRD/" + uk_labels["fileName"]

        metadata = pd.concat([uk_labels, ec_labels])
        metadata = metadata[[(self.base_dir / fp).exists() for fp in metadata["filePath"]]]
        # metadata = self._label_encode_metadata(metadata)
        metadata.to_csv(self._metadata_path, index=False)

    def _preprocess_files(self) -> None:
        """
        Segment the waveform files based on :py:attr:`segment_len` and cache the file-segments.
        """
        if self.segment_len is None:
            return
        processed_audio_dir = self.base_dir / f"segment_len={self.segment_len}"
        processed_audio_dir.mkdir(parents=True, exist_ok=True)

        waveform_paths = self.base_dir / self.metadata["filePath"]  # type: ignore
        segment_filenames: List[Tuple[str, str]] = []
        for path in tqdm(waveform_paths, desc="Preprocessing", colour=self._PBAR_COL):
            waveform_filename = path.stem
            waveform, sr = torchaudio.load(path)  # type: ignore
            audio_len = waveform.size(-1) / sr
            frac_remainder, num_segments = math.modf(audio_len / self.segment_len)
            num_segments = int(num_segments)

            if frac_remainder >= 0.5:
                self.logger.debug(
                    (
                        f"Length of audio file '{path.resolve()}' is not integer-divisible by "
                        f"{self.segment_len}: terminally zero-padding the file along the "
                        "time-axis to compensate."
                    ),
                )
                padding = torch.zeros(
                    waveform.size(0),
                    int((self.segment_len - (frac_remainder * self.segment_len)) * sr),
                )
                waveform = torch.cat((waveform, padding), dim=-1)
                num_segments += 1
            if 0 < frac_remainder < 0.5:
                self.logger.debug(
                    (
                        f"Length of audio file '{path.resolve()}' is not integer-divisible by"
                        f" {self.segment_len} and not of sufficient length to be padded (fractional"
                        " remainder must be greater than 0.5): discarding terminal segment."
                    ),
                )
                waveform = waveform[:, : int(num_segments * self.segment_len * sr)]

            waveform_segments = waveform.chunk(chunks=num_segments, dim=-1)
            for seg_idx, segment in enumerate(waveform_segments):
                segment_filename = f"{waveform_filename}_{seg_idx}.wav"
                segment_filepath = processed_audio_dir / segment_filename
                torchaudio.save(  # type: ignore
                    segment_filepath,
                    src=segment,
                    sample_rate=sr,
                )
                segment_filenames.append(
                    (waveform_filename, str(segment_filepath.relative_to(self.base_dir)))
                )

        pd.DataFrame(segment_filenames, columns=["fileName", "filePath"]).to_csv(
            processed_audio_dir / "filepaths.csv", index=False
        )

    @override
    def load_sample(self, index: IndexType) -> Tensor:
        assert isinstance(index, int)
        path = self.audio_dir / self.x[index]

        # get metadata first
        metadata = torchaudio.info(path)  # type: ignore

        # compute number of frames to take with the real sample rate
        num_frames_segment = int(
            self.num_frames_in_segment / self.sample_rate * metadata.sample_rate
        )

        # get random sub-sample
        high = max(1, metadata.num_frames - num_frames_segment)
        frame_offset: int = torch.randint(low=0, high=high, size=(1,)).item()  # type: ignore

        # load segment
        waveform, _ = torchaudio.load(  # type: ignore
            path, num_frames=num_frames_segment, frame_offset=frame_offset
        )

        # resample to correct sample rate
        waveform = torchaudio.functional.resample(
            waveform,
            orig_freq=metadata.sample_rate,
            new_freq=self.sample_rate,
        )

        return waveform



"""NICO Dataset."""

from enum import Enum, auto
from functools import cached_property
import json
from pathlib import Path
import random
from typing import ClassVar, Dict, List, Literal, Optional, Sequence, Set, Tuple, Union
from typing_extensions import TypeAlias

import pandas as pd
from ranzen import StrEnum
import torch
from torch import Tensor

from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["NICOPP", "NicoPPTarget", "NicoPPAttr", "NicoPPSplit"]


class NicoPPTarget(StrEnum):
    """Class labels in NICO++."""

    AIRPLANE = auto()
    BEAR = auto()
    BICYCLE = auto()
    BIRD = auto()
    BUS = auto()
    BUTTERFLY = auto()
    CACTUS = auto()
    CAR = auto()
    CAT = auto()
    CHAIR = auto()
    CLOCK = auto()
    CORN = auto()
    COW = auto()
    CRAB = auto()
    CROCODILE = auto()
    DOG = auto()
    DOLPHIN = auto()
    ELEPHANT = auto()
    FISHING_ROD = "fishing rod"  # class name contains space
    FLOWER = auto()
    FOOTBALL = auto()
    FOX = auto()
    FROG = auto()
    GIRAFFE = auto()
    GOOSE = auto()
    GUN = auto()
    HAT = auto()
    HELICOPTER = auto()
    HORSE = auto()
    HOT_AIR_BALLOON = "hot air balloon"  # class name contains space
    KANGAROO = auto()
    LIFEBOAT = auto()
    LION = auto()
    LIZARD = auto()
    MAILBOX = auto()
    MONKEY = auto()
    MOTORCYCLE = auto()
    OSTRICH = auto()
    OWL = auto()
    PINEAPPLE = auto()
    PUMPKIN = auto()
    RABBIT = auto()
    RACKET = auto()
    SAILBOAT = auto()
    SCOOTER = auto()
    SEAL = auto()
    SHEEP = auto()
    SHIP = auto()
    SHRIMP = auto()
    SPIDER = auto()
    SQUIRREL = auto()
    SUNFLOWER = auto()
    TENT = auto()
    TIGER = auto()
    TORTOISE = auto()
    TRAIN = auto()
    TRUCK = auto()
    UMBRELLA = auto()
    WHEAT = auto()
    WOLF = auto()


class NicoPPAttr(StrEnum):
    """Common attributes in NICO++."""

    AUTUMN = auto()
    DIM = auto()
    GRASS = auto()
    OUTDOOR = auto()
    ROCK = auto()
    WATER = auto()


class NicoPPSplit(Enum):
    TRAIN = 0
    VAL = 1
    TEST = 2


SampleType: TypeAlias = TernarySample


class NICOPP(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """NICO++ dataset."""

    SampleType: TypeAlias = TernarySample
    Superclass: TypeAlias = NicoPPTarget
    Subclass: TypeAlias = NicoPPAttr
    Split: TypeAlias = NicoPPSplit

    data_split_seed: ClassVar[int] = 666  # this is the seed from the paper
    num_samples_val_test: ClassVar[int] = 75  # this is the number from the paper
    subpath: ClassVar[Path] = Path("public_dg_0416") / "train"

    def __init__(
        self,
        root: Union[str, Path],
        *,
        transform: Optional[ImageTform] = None,
        superclasses: Optional[Sequence[Union[NicoPPTarget, str]]] = None,
        split: Optional[Union[NicoPPSplit, str]] = None,
    ) -> None:
        self.superclasses: Optional[List[NicoPPTarget]] = None
        if superclasses is not None:
            assert superclasses, "superclasses should be a non-empty list"
            self.superclasses = [NicoPPTarget(superclass) for superclass in superclasses]
        self.split = NicoPPSplit[split.upper()] if isinstance(split, str) else split

        self.root = Path(root)
        self._base_dir = self.root / "nico_plus_plus"
        self._metadata_path = self._base_dir / "metadata.csv"

        if not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )
        if not self._metadata_path.exists():
            self._extract_metadata()

        self.metadata = pd.read_csv(self._metadata_path)

        if self.superclasses is not None:
            self.metadata = self.metadata[self.metadata["superclass"].isin(self.superclasses)]

        if self.split is not None:
            self.metadata = self.metadata[self.metadata["split"] == self.split.value]

        # Divide up the dataframe into its constituent arrays because indexing with pandas is
        # substantially slower than indexing with numpy/torch
        x = self.metadata["filename"].to_numpy()
        y = torch.as_tensor(self.metadata["y"].to_numpy(), dtype=torch.long)
        s = torch.as_tensor(self.metadata["a"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)

    @cached_property
    def class_tree(self) -> Dict[str, Set[str]]:
        return self.metadata[["y", "a"]].drop_duplicates().groupby("y").agg(set).to_dict()["a"]

    @cached_property
    def superclass_label_decoder(self) -> Dict[int, NicoPPTarget]:
        return dict((val, NicoPPTarget(name)) for name, val in self._get_label_mapping("y"))

    @cached_property
    def subclass_label_decoder(self) -> Dict[int, NicoPPAttr]:
        return dict((val, NicoPPAttr(name)) for name, val in self._get_label_mapping("a"))

    def _get_label_mapping(self, level: Literal["y", "a"]) -> List[Tuple[str, int]]:
        """Get a list of all possible (name, numerical value) pairs."""
        return [
            (name, num)
            for num, name in self.metadata[[f"{level}_name", level]]
            .drop_duplicates()
            .itertuples(index=False, name=None)
        ]

    def _check_unzipped(self) -> bool:
        if not all((self._base_dir / self.subpath / attr).exists() for attr in NicoPPAttr):
            return False
        if not (self._base_dir / "dg_label_id_mapping.json").exists():
            return False
        return True

    def _extract_metadata(self) -> None:
        self.logger.info("Generating metadata for NICO++...")
        attributes = ["autumn", "dim", "grass", "outdoor", "rock", "water"]  # 6 attrs, 60 labels
        meta = json.load((self._base_dir / "dg_label_id_mapping.json").open("r"))

        def _make_balanced_testset(
            df: pd.DataFrame, *, seed: int, num_samples_val_test: int
        ) -> pd.DataFrame:
            # each group has a test set size of (2/3 * num_samples_val_test) and a val set size of
            # (1/3 * num_samples_val_test); if total samples in original group < num_samples_val_test,
            # val/test will still be split by 1:2, but no training samples remained

            random.seed(seed)
            val_set, test_set = [], []
            for g in pd.unique(df["g"]):
                df_group = df[df["g"] == g]
                curr_data = df_group["filename"].values
                random.shuffle(curr_data)
                split_size = min(len(curr_data), num_samples_val_test)
                val_set += list(curr_data[: split_size // 3])
                test_set += list(curr_data[split_size // 3 : split_size])
            self.logger.info(f"Val: {len(val_set)}, Test: {len(test_set)}")
            assert len(set(val_set).intersection(set(test_set))) == 0
            combined_set = dict(zip(val_set, [NicoPPSplit.VAL.value for _ in range(len(val_set))]))
            combined_set.update(
                dict(zip(test_set, [NicoPPSplit.TEST.value for _ in range(len(test_set))]))
            )
            df["split"] = df["filename"].map(combined_set)
            df["split"].fillna(NicoPPSplit.TRAIN.value, inplace=True)
            df["split"] = df.split.astype(int)
            return df

        all_data = []
        for c, attr in enumerate(attributes):
            for label in meta:
                folder_path = self._base_dir / self.subpath / attr / label
                y = meta[label]
                for img_path in Path(folder_path).glob("*.jpg"):
                    all_data.append(
                        {
                            "filename": str(img_path.relative_to(self._base_dir)),
                            "y": y,
                            "a": c,
                            "y_name": label,
                            "a_name": attr,
                        }
                    )
        df = pd.DataFrame(all_data)
        df["g"] = df["a"] + df["y"] * len(attributes)
        df = _make_balanced_testset(
            df, seed=self.data_split_seed, num_samples_val_test=self.num_samples_val_test
        )
        df = df.drop(columns=["g"])
        df.to_csv(self._metadata_path, index=False)



"""Datasets from EthicML."""

from itertools import groupby
from typing import Iterator, List, Tuple

from ethicml import DataTuple
from ethicml.implementations.pytorch_common import _get_info
import torch

from conduit.data.datasets.base import CdtDataset

__all__ = ["DataTupleDataset"]


def group_features(disc_feats: List[str]) -> Iterator[Tuple[str, Iterator[str]]]:
    """Group discrete features names according to the first segment of their name."""

    def _first_segment(feature_name: str) -> str:
        return feature_name.split("_")[0]

    return groupby(disc_feats, _first_segment)


def grouped_features_indexes(disc_feats: List[str]) -> List[slice]:
    """Group discrete features names according to the first segment of their name.

    Then return a list of their corresponding slices (assumes order is maintained).
    """
    group_iter = group_features(disc_feats)

    feature_slices = []
    start_idx = 0
    for _, group in group_iter:
        len_group = len(list(group))
        indexes = slice(start_idx, start_idx + len_group)
        feature_slices.append(indexes)
        start_idx += len_group

    return feature_slices


class DataTupleDataset(CdtDataset):
    """Wrapper for EthicML datasets."""

    def __init__(self, dataset: DataTuple, disc_features: List[str], cont_features: List[str]):
        """Create DataTupleDataset."""
        disc_features = [feat for feat in disc_features if feat in dataset.x.columns]
        self.disc_features = disc_features

        cont_features = [feat for feat in cont_features if feat in dataset.x.columns]
        self.cont_features = cont_features
        self.feature_groups = dict(discrete=grouped_features_indexes(self.disc_features))

        self.x_disc = torch.as_tensor(dataset.x[self.disc_features].to_numpy(), dtype=torch.long)
        self.x_cont = torch.as_tensor(dataset.x[self.cont_features].to_numpy(), dtype=torch.float)
        x = torch.cat([self.x_disc, self.x_cont], dim=1)

        # NOTE: we should probably not use an internal function from EthicML here
        _, s, self.num, self.xdim, self.x_names, self.s_name = _get_info(dataset)
        self.sdim = 1

        y = torch.as_tensor(dataset.y.to_numpy(), dtype=torch.float32)

        self.ydim = 1
        self.y_name = dataset.y_column

        super().__init__(x=x, y=y, s=s)



from collections.abc import Iterable
from enum import Enum, auto
from typing import TypeAlias
from typing_extensions import Self, TypeAliasType

from folktables import ACSDataSource, ACSEmployment, ACSIncome, BasicProblem, generate_categories
import numpy as np
import pandas as pd
from torch import Tensor

from conduit.data.datasets import random_split, stratified_split
from conduit.data.datasets.tabular.base import CdtTabularDataset
from conduit.data.structures import TernarySample

__all__ = ["ACSSetting", "ACSDataset", "ACSState", "ACSHorizon", "ACSSurvey", "ACSSurveyYear"]


class ACSSurveyYear(Enum):
    YEAR_2014 = "2014"
    YEAR_2015 = "2015"
    YEAR_2016 = "2016"
    YEAR_2017 = "2017"
    YEAR_2018 = "2018"


class ACSState(Enum):
    AL = auto()
    AK = auto()
    AZ = auto()
    AR = auto()
    CA = auto()
    CO = auto()
    CT = auto()
    DE = auto()
    FL = auto()
    GA = auto()
    HI = auto()
    ID = auto()
    IL = auto()
    IN = auto()
    IA = auto()
    KS = auto()
    KY = auto()
    LA = auto()
    ME = auto()
    MD = auto()
    MA = auto()
    MI = auto()
    MN = auto()
    MS = auto()
    MO = auto()
    MT = auto()
    NE = auto()
    NV = auto()
    NH = auto()
    NJ = auto()
    NM = auto()
    NY = auto()
    NC = auto()
    ND = auto()
    OH = auto()
    OK = auto()
    OR = auto()
    PA = auto()
    RI = auto()
    SC = auto()
    SD = auto()
    TN = auto()
    TX = auto()
    UT = auto()
    VT = auto()
    VA = auto()
    WA = auto()
    WV = auto()
    WI = auto()
    WY = auto()
    PR = auto()


class ACSHorizon(Enum):
    ONE_YEAR = "1-Year"
    FIVE_YEARS = "5-Year"


class ACSSurvey(Enum):
    PERSON = "person"
    HOUSEHOLD = "household"


def _employment_filter(data: pd.DataFrame) -> pd.DataFrame:
    """Custom filter.

    (Age) must be greater than 16 and less than 90, and (Person weight) must be
    greater than or equal to 1.
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['AGEP'] < 90]
    df = df[df['PWGTP'] >= 1]
    return df


_ACSEmploymentDisability = BasicProblem(
    features=[
        'AGEP',  # age; for range of values of features please check Appendix B.4 of Retiring Adult:
        # New Datasets for Fair Machine Learning NeurIPS 2021 paper
        'SCHL',  # educational attainment
        'MAR',  # marital status
        'RELP',  # relationship
        'DIS',  # disability recode
        'ESP',  # employment status of parents
        'CIT',  # citizenship status
        'MIG',  # mobility status (lived here 1 year ago)
        'MIL',  # military service
        'ANC',  # ancestry recode
        'NATIVITY',  # nativity
        'DEAR',  # hearing difficulty
        'DEYE',  # vision difficulty
        'DREM',  # cognitive difficulty
        'SEX',  # sex
        'RAC1P',  # recoded detailed race code
        'GCL',  # grandparents living with grandchildren
    ],
    target='ESR',  # employment status recode
    target_transform=lambda x: x == 1,
    group='DIS',
    preprocess=_employment_filter,
    postprocess=lambda x: np.nan_to_num(x, nan=-1),
)


class ACSSetting(Enum):
    employment = ACSEmployment
    income = ACSIncome
    employment_disability = _ACSEmploymentDisability


class ACSDataset(CdtTabularDataset[TernarySample, Tensor, Tensor]):
    """Wrapper for the ACS dataset from Folktables."""

    Setting: TypeAlias = ACSSetting
    Horizon: TypeAlias = ACSHorizon
    State: TypeAlias = ACSState
    Survey: TypeAlias = ACSSurvey
    SurveyYear: TypeAlias = ACSSurveyYear

    def __init__(
        self,
        setting: ACSSetting,
        survey_year: SurveyYear = SurveyYear.YEAR_2018,
        horizon: Horizon = Horizon.ONE_YEAR,
        survey: ACSSurvey = ACSSurvey.PERSON,
        states: Iterable[ACSState] = (ACSState.AL,),
    ):
        data_source = ACSDataSource(
            survey_year=survey_year.value, horizon=horizon.value, survey=survey.value
        )
        acs_data = data_source.get_data(states=[state.name for state in states], download=True)
        dataset: BasicProblem = setting.value

        # `generate_categories` is only available for years >= 2017.
        if int(survey_year.value) >= 2017:
            categories = generate_categories(
                features=dataset.features,
                definition_df=data_source.get_definitions(download=True),
            )

            # One-hot encoding based on the categories.
            features_df, label_df, group_df = dataset.df_to_pandas(
                acs_data, categories=categories, dummies=True
            )

            feature_groups, ohe_indexes = feature_groups_from_categories(categories, features_df)
            label = label_df.to_numpy(dtype=np.int64)
            group = group_df.to_numpy(dtype=np.int64)
            features = features_df.to_numpy(dtype=np.float32)
            non_ohe_indexes = list(set(range(features.shape[1])) - set(ohe_indexes))
        else:
            # Categorical features are *not* one-hot encoded for years < 2017.
            features, label, group = dataset.df_to_numpy(acs_data)
            non_ohe_indexes, feature_groups = None, None

        super().__init__(
            x=features,
            y=label,
            s=group,
            non_ohe_indexes=non_ohe_indexes,
            feature_groups=feature_groups,
        )

    def subsampled_split(
        self,
        train_props: dict[int, dict[int, float]] | None,
        *,
        val_prop: float,
        test_prop: float,
        seed: int,
    ) -> tuple[Self, Self, Self]:
        """Create a split of the dataset in which the training set is missing certain groups.

        :param train_props: Specification for which groups to drop from the training set.
        :param val_prop: Proportion of the data to use for the validation set.
        :param test_prop: Proportion of the data to use for the test set.
        :param seed: PRNG seed to use for splitting the data.
        :returns: Random subsets of the data of the requested proportions.
        """
        val_test_prop = val_prop + test_prop
        train_data, val_test_data = stratified_split(
            self,
            default_train_prop=1 - val_test_prop,
            train_props=train_props,
            seed=seed,
            reproducible=True,
        )
        val_data, test_data = random_split(
            val_test_data, props=val_prop / val_test_prop, seed=seed, reproducible=True
        )
        return train_data, val_data, test_data


CategoryName = TypeAliasType("CategoryName", str)
ValueName = TypeAliasType("ValueName", str)


def feature_groups_from_categories(
    categories: dict[CategoryName, dict[float, ValueName]], features: pd.DataFrame
) -> tuple[list[slice], list[int]]:
    slices: list[slice] = []
    disc_indexes: list[int] = []

    for category_name, value_entries in categories.items():
        indexes = []
        for value_name in value_entries.values():
            feature_name = f"{category_name}_{value_name}"
            if feature_name in features.columns:
                feature_index = features.columns.get_loc(feature_name)
                indexes.append(feature_index)

        # Determine the slice bounds for this category.
        start = min(indexes)
        stop = max(indexes) + 1

        # Check that the indexes are contiguous.
        index_set = set(indexes)
        for i in range(start, stop):
            assert i in index_set

        slices.append(slice(start, stop))
        disc_indexes.extend(indexes)
    return slices, disc_indexes



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from typing import Any


@dataclass
class DecoupledContrastiveLossConf:
    _target_: str = "conduit.models.self_supervised.loss.DecoupledContrastiveLoss"
    temperature: float = 0.1
    weight_fn: Any = None  # Optional[Callable[[Tensor, Tensor], Tensor]]



from enum import Enum, auto
from pathlib import Path
from typing import ClassVar, Optional, Union
from typing_extensions import TypeAlias

import pandas as pd
from ranzen.misc import StrEnum
import torch
from torch import Tensor

from conduit.data.datasets import UrlFileInfo, download_from_url
from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["Camelyon17", "Camelyon17Split", "Camelyon17SplitScheme"]


class Camelyon17SplitScheme(StrEnum):
    OFFICIAL = auto()
    """Oficial split."""

    MIXED_TO_TEST = auto()
    """
    For the mixed-to-test setting, slide 23 (corresponding to patient 042, node 3 in the
    original dataset) is moved from the test set to the training set
    """


class Camelyon17Split(Enum):
    TRAIN = 0
    ID_VAL = 1
    TEST = 2
    VAL = 3


class Camelyon17Attr(StrEnum):
    TUMOR = auto()
    CENTER = auto()
    SLIDE = auto()


SampleType: TypeAlias = TernarySample


class Camelyon17(CdtVisionDataset[SampleType, Tensor, Tensor]):
    """The CAMELYON17-WILDS histopathology dataset.

    This is a modified version of the original CAMELYON17 dataset.

    Supported ``split_scheme``:
        - 'official'
        - 'mixed-to-test'

    Input (x):
        96x96 image patches extracted from histopathology slides.

    Label (y):
        y is binary. It is 1 if the central 32x32 region contains any tumor tissue, and 0 otherwise.

    Metadata:
        Each patch is annotated with the ID of the hospital it came from (integer from 0 to 4)
        and the slide it came from (integer from 0 to 49).

    Website:
        https://camelyon17.grand-challenge.org/

    Original publication:

    .. code-block:: bibtex

        @article{bandi2018detection,
          title={From detection of individual metastases to classification of lymph node status at
                 the patient level: the camelyon17 challenge},
          author={Bandi, Peter and Geessink, Oscar and Manson, Quirine and Van Dijk, Marcory
                  and Balkenhol, Maschenka and Hermsen, Meyke and Bejnordi, Babak Ehteshami
                  and Lee, Byungjae and Paeng, Kyunghyun and Zhong, Aoxiao and others},
          journal={IEEE transactions on medical imaging},
          volume={38},
          number={2},
          pages={550--560},
          year={2018},
          publisher={IEEE}
        }

    License:
        This dataset is in the public domain and is distributed under CC0.
        https://creativecommons.org/publicdomain/zero/1.0/
    """

    SampleType: TypeAlias = TernarySample
    Attr: TypeAlias = Camelyon17Attr
    Split: TypeAlias = Camelyon17Split
    SplitScheme: TypeAlias = Camelyon17SplitScheme

    _FILE_INFO: ClassVar[UrlFileInfo] = UrlFileInfo(
        name="camelyon17_v1.0.tar.gz",
        url="https://worksheets.codalab.org/rest/bundles/0xe45e15f39fb54e9d9e919556af67aabe/contents/blob/",
        md5=None,
    )
    _TEST_CENTER: ClassVar[int] = 2
    _VAL_CENTER: ClassVar[int] = 1

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        split: Optional[Union[Camelyon17Split, str]] = None,
        split_scheme: Union[Camelyon17SplitScheme, str] = Camelyon17SplitScheme.OFFICIAL,
        superclass: Union[Camelyon17Attr, str] = Camelyon17Attr.TUMOR,
        subclass: Union[Camelyon17Attr, str] = Camelyon17Attr.CENTER,
    ) -> None:
        self.superclass = Camelyon17Attr(superclass) if isinstance(split, str) else split
        self.subclass = Camelyon17Attr(subclass) if isinstance(split, str) else split
        self.split = Camelyon17Split[split.upper()] if isinstance(split, str) else split
        self.split_scheme = (
            Camelyon17SplitScheme(split_scheme) if isinstance(split_scheme, str) else split_scheme
        )

        self.root = Path(root)
        self._base_dir = self.root / "camelyon17_v1.0"
        self.download = download
        if self.download:
            download_from_url(
                file_info=self._FILE_INFO,
                root=self.root,
                logger=self.logger,
                remove_finished=True,
            )
        else:
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )

        # Read in metadata
        # Note: metadata is one-indexed.
        self.metadata = pd.read_csv(
            self._base_dir / 'metadata.csv', index_col=0, dtype={"patient": "str"}
        )
        if self.split_scheme is Camelyon17SplitScheme.MIXED_TO_TEST:
            # For the mixed-to-test setting,
            # we move slide 23 (corresponding to patient 042, node 3 in the original dataset)
            # from the test set to the training set
            slide_mask = self.metadata["slide"] == 23
            self.metadata.loc[slide_mask, "split"] = Camelyon17Split.TRAIN.value
        # Use an official split of the data, if 'split' is specified, else just use all
        # of the data
        val_center_mask = self.metadata["center"] == self._VAL_CENTER
        test_center_mask = self.metadata["center"] == self._TEST_CENTER
        self.metadata.loc[val_center_mask, "split"] = Camelyon17Split.VAL.value
        self.metadata.loc[test_center_mask, "split"] = Camelyon17Split.TEST.value

        if self.split is not None:
            split_indices = self.metadata["split"] == self.split.value
            self.metadata = self.metadata.loc[split_indices]

        # Construct filepaths from metadata
        def build_fp(row: pd.DataFrame) -> str:
            return "patches/patient_{0}_node_{1}/patch_patient_{0}_node_{1}_x_{2}_y_{3}.png".format(
                *row
            )

        x = (
            self.metadata[["patient", "node", "x_coord", "y_coord"]]
            .apply(build_fp, axis=1)
            .to_numpy()
        )
        # Extract superclass labels
        y = torch.as_tensor(self.metadata[str(self.superclass)].to_numpy(), dtype=torch.long)
        # Extract subclass labels
        s = torch.as_tensor(self.metadata[str(self.subclass)].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)



from dataclasses import fields, is_dataclass
from functools import lru_cache
import logging
from multiprocessing.context import BaseContext
from pathlib import Path
import platform
import subprocess
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterator,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    TypedDict,
    Union,
    cast,
    overload,
)
from typing_extensions import TypeAlias, TypeGuard, Unpack
from zipfile import BadZipFile

import numpy as np
import numpy.typing as npt
from ranzen.misc import gcopy
from ranzen.torch.data import Subset, prop_random_split
import torch
from torch import Tensor
from torch.utils.data import ConcatDataset
from torch.utils.data._utils.collate import default_collate_err_msg_format, np_str_obj_array_pattern
from torch.utils.data.dataloader import DataLoader, _worker_init_fn_t
from torch.utils.data.sampler import Sampler
from torchvision.datasets.utils import (  # type: ignore
    _detect_file_type,
    download_url,
    extract_archive,
)

from conduit.data.datasets.base import CdtDataset
from conduit.data.structures import (
    BinarySample,
    Dataset,
    LoadedData,
    NamedSample,
    PseudoCdtDataset,
    SampleBase,
    SizedDataset,
    TernarySample,
    TrainTestSplit,
)

__all__ = [
    "AudioTform",
    "CdtDataLoader",
    "GdriveFileInfo",
    "UrlFileInfo",
    "cdt_collate",
    "check_integrity",
    "download_from_gdrive",
    "download_from_url",
    "extract_base_dataset",
    "extract_labels_from_dataset",
    "get_group_ids",
    "infer_al_backend",
    "infer_sample_cls",
    "is_tensor_list",
    "make_subset",
    "random_split",
    "stratified_split",
]


AudioLoadingBackend: TypeAlias = Literal["sox_io", "soundfile"]


def infer_al_backend() -> AudioLoadingBackend:
    """Infer which audio-loading backend to use based on the operating system."""
    soundfile: Final = "soundfile"
    sox: Final = "sox_io"
    return soundfile if platform.system() == "Windows" else sox


AudioTform: TypeAlias = Callable[[Tensor], Tensor]


def apply_audio_transform(waveform: Tensor, *, transform: Optional[AudioTform]) -> Tensor:
    return waveform if transform is None else transform(waveform)


@overload
def extract_base_dataset(
    dataset: Dataset, *, return_subset_indices: Literal[True] = ...
) -> Tuple[Dataset, Union[Tensor, slice]]:
    ...


@overload
def extract_base_dataset(
    dataset: Dataset, *, return_subset_indices: Literal[False] = ...
) -> Dataset:
    ...


def extract_base_dataset(
    dataset: Dataset, *, return_subset_indices: bool = True
) -> Union[Dataset, Tuple[Dataset, Union[Tensor, slice]]]:
    """Extract the innermost dataset of a nesting of datasets.

    Nested datasets are inferred based on the existence of a 'dataset'
    attribute and the base dataset is extracted by recursive application
    of this rule.

    :param dataset: The dataset from which to extract the base dataset.

    :param return_subset_indices: Whether to return the indices from which the overall subset of the
        dataset was created (works for multiple levels of subsetting).

    :returns: The base dataset, which may be the original dataset if one does not exist or cannot be
        determined.
    """

    def _closure(
        dataset: Dataset, rel_indices_ls: Optional[List[List[int]]] = None
    ) -> Union[Dataset, Tuple[Dataset, Union[Tensor, slice]]]:
        if rel_indices_ls is None:
            rel_indices_ls = []
        if hasattr(dataset, "dataset"):
            if isinstance(dataset, Subset):
                rel_indices_ls.append(list(dataset.indices))
            return _closure(dataset.dataset, rel_indices_ls)  # type: ignore
        if return_subset_indices:
            if rel_indices_ls:
                abs_indices = torch.as_tensor(rel_indices_ls.pop(), dtype=torch.long)
                for indices in rel_indices_ls[::-1]:
                    abs_indices = abs_indices[indices]
            else:
                abs_indices = slice(None)
            return dataset, abs_indices
        return dataset

    return _closure(dataset)


@lru_cache(typed=True)
def extract_labels_from_dataset(
    dataset: PseudoCdtDataset,
) -> Tuple[Optional[Tensor], Optional[Tensor]]:
    """Attempt to extract s/y labels from a dataset."""
    base_dataset = dataset

    def _closure(dataset: Dataset) -> Tuple[Optional[Tensor], Optional[Tensor]]:
        base_dataset, indices = extract_base_dataset(dataset=dataset, return_subset_indices=True)
        _s = None
        _y = None
        if (s := getattr(base_dataset, "s", None)) is not None:
            _s = s[indices]
        if (y := getattr(base_dataset, "y", None)) is not None:
            _y = y[indices]

        _s = torch.from_numpy(_s) if isinstance(_s, np.ndarray) else _s
        _y = torch.from_numpy(_y) if isinstance(_y, np.ndarray) else _y

        return _s, _y

    if isinstance(base_dataset, (ConcatDataset)):
        s_all_ls: List[Tensor] = []
        y_all_ls: List[Tensor] = []
        for _dataset in base_dataset.datasets:
            s, y = _closure(_dataset)
            if s is not None:
                s_all_ls.append(s)
            if y is not None:
                s_all_ls.append(y)
        s_all = torch.cat(s_all_ls, dim=0) if s_all_ls else None
        y_all = torch.cat(y_all_ls, dim=0) if y_all_ls else None
    else:
        s_all, y_all = _closure(base_dataset)
    return s_all, y_all


def get_group_ids(dataset: Dataset) -> Tensor:
    s_all, y_all = extract_labels_from_dataset(dataset)
    # group_ids: Optional[Tensor] = None
    if s_all is None:
        if y_all is None:
            raise ValueError(
                "Unable to compute group ids for dataset because no labels could be extracted."
            )
        group_ids = y_all
    elif y_all is None:
        group_ids = s_all
    else:
        group_ids = (y_all * len(s_all.unique()) + s_all).squeeze()
    return group_ids.long()


def compute_instance_weights(dataset: Dataset, *, upweight: bool = False) -> Tensor:
    group_ids = get_group_ids(dataset)
    _, inv_indexes, counts = group_ids.unique(return_inverse=True, return_counts=True)
    # Upweight samples according to the cardinality of their intersectional group
    if upweight:
        group_weights = len(group_ids) / counts
    # Downweight samples according to the cardinality of their intersectional group
    # - this approach should be preferred due to being more numerically stable
    # (very small counts can lead to very large weighted loss values when upweighting)
    else:
        counts_r = counts.reciprocal()
        group_weights = counts_r / counts_r.sum()
    return group_weights[inv_indexes]


PCD = TypeVar("PCD", bound=PseudoCdtDataset)


def make_subset(
    dataset: Union[PCD, Subset[PCD]],
    *,
    indices: Optional[Union[List[int], npt.NDArray[np.uint64], Tensor, slice]],
    deep: bool = False,
) -> PCD:
    """Create a subset of the dataset from the given indices.

    :param dataset: The dataset to split.
    :param indices: The sample-indices from which to create the subset.
        In the case of being a numpy array or tensor, said array or tensor
        must be 0- or 1-dimensional.

    :param deep: Whether to create a copy of the underlying dataset as a basis for the subset.
        If False then the data of the subset will be a view of original dataset's data.

    :returns: A subset of the dataset from the given indices.
    :raises ValueError: if the indices don't have the correct shape.
    :raises TypeError: if the dataset is not an instance of ``CdtDataset``.
    """
    if isinstance(indices, (np.ndarray, Tensor)):
        if indices.ndim > 1:
            raise ValueError("If 'indices' is an array it must be a 0- or 1-dimensional.")
        indices = cast(List[int], indices.tolist())

    current_indices = None
    if isinstance(dataset, Subset):
        base_dataset, current_indices = extract_base_dataset(dataset, return_subset_indices=True)
        if not isinstance(base_dataset, CdtDataset):
            raise TypeError(
                f"Subsets can only be created from {CdtDataset.__name__} instances or PyTorch "
                "Subsets of them."
            )
        base_dataset = cast(PCD, base_dataset)

        if isinstance(current_indices, Tensor):
            current_indices = current_indices.tolist()
    else:
        base_dataset = dataset
    subset = gcopy(base_dataset, deep=deep)

    def _subset_from_indices(_dataset: PCD, _indices: Union[List[int], slice]) -> PCD:
        _dataset.x = _dataset.x[_indices]
        if _dataset.y is not None:
            _dataset.y = _dataset.y[_indices]
        if _dataset.s is not None:
            _dataset.s = _dataset.s[_indices]
        return _dataset

    if current_indices is not None:
        subset = _subset_from_indices(_dataset=subset, _indices=current_indices)
    if indices is not None:
        subset = _subset_from_indices(_dataset=subset, _indices=indices)

    return subset


def infer_sample_cls(
    sample: Union[List[LoadedData], Tuple[LoadedData, ...], Dict[str, LoadedData], LoadedData],
) -> Type[SampleBase]:
    """ "Attempt to infer the appropriate sample class based on the length of the input."""
    if not isinstance(sample, (list, tuple, dict)) or (len(sample) == 1):
        return NamedSample
    elif len(sample) == 2:
        return BinarySample
    elif len(sample) == 3:
        return TernarySample
    else:
        raise ValueError("Only items with 3 or fewer elements can be cast to 'Sample' instances.")


class cdt_collate:
    def __init__(
        self, cast_to_sample: bool = True, *, converter: Optional[Union[Type[Any], Callable]] = None
    ) -> None:
        self.cast_to_sample = cast_to_sample
        self.converter = converter

    def _collate(self, batch: Any) -> Any:
        elem = batch[0]
        elem_type = type(elem)
        # dataclass
        if is_dataclass(elem):
            return elem_type(
                **{
                    field.name: self._collate([getattr(d, field.name) for d in batch])
                    for field in fields(elem)
                }
            )
        # Tensor
        elif isinstance(elem, Tensor):
            out = None
            if torch.utils.data.get_worker_info() is not None:  # type: ignore
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum(x.numel() for x in batch)
                if hasattr(elem, "_typed_storage"):  # in PyTorch 2.0, it's called _typed_storage
                    storage = elem._typed_storage()._new_shared(numel, device=elem.device)
                else:
                    storage = elem.storage()._new_shared(numel)
                out = elem.new(storage).resize_(len(batch), *list(elem.size()))
            ndims = elem.dim()
            # If 'batch' is a sequence of sub-batched tensors we concatenate the elements along the
            # batch dimesion. Note, the problem of inferring whether the tensors are sub-batched or
            # not is ill-posed given that the dimensionality of samples varies with modality; the
            # current solution is tailored for tabular and image data.
            if (ndims > 0) and ((ndims % 2) == 0):
                return torch.cat(batch, dim=0, out=out)
            return torch.stack(batch, dim=0, out=out)
        # NumPy array
        elif (
            elem_type.__module__ == "numpy"
            and elem_type.__name__ != "str_"
            and elem_type.__name__ != "string_"
        ):
            elem = batch[0]
            if elem_type.__name__ == "ndarray":
                # array of string classes and object
                if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                    raise TypeError(default_collate_err_msg_format.format(elem.dtype))
                return self._collate([torch.as_tensor(b) for b in batch])
        # Float
        elif isinstance(elem, float):
            return torch.tensor(batch, dtype=torch.float32)
        # Integer
        elif isinstance(elem, int):
            return torch.tensor(batch)
        # String
        elif isinstance(elem, str):
            return batch
        # Mapping
        elif isinstance(elem, Mapping):
            return elem_type({key: self._collate([d[key] for d in batch]) for key in elem})
        # NamedTuple
        elif isinstance(elem, tuple) and hasattr(elem, "_fields"):  # namedtuple
            return elem_type(**(self._collate(samples) for samples in zip(*batch)))
        # Tuple or List
        elif isinstance(elem, (tuple, list)):
            transposed = zip(*batch)
            return elem_type([self._collate(samples) for samples in transposed])
        # Invalid (uncollatable) type
        raise TypeError(default_collate_err_msg_format.format(elem_type))

    def __call__(self, batch: Any) -> Any:
        collated_batch = self._collate(batch=batch)
        if self.converter is not None:
            collated_batch = self.converter(collated_batch)
        if self.cast_to_sample and (not isinstance(collated_batch, SampleBase)):
            if isinstance(collated_batch, Tensor):
                collated_batch = NamedSample(x=collated_batch)
            elif isinstance(collated_batch, (tuple, list, dict)):
                sample_cls = infer_sample_cls(collated_batch)
                if not isinstance(collated_batch, dict):
                    collated_batch = dict(zip(["x", "y", "s"], collated_batch))
                collated_batch = sample_cls(**collated_batch)
            else:
                raise ValueError(
                    f"batch of type '{type(collated_batch)}' could not be automatically cast to a "
                    "'Sample' instance. Batch must be of type 'dict', 'tuple', or 'list'."
                )
        return collated_batch


I = TypeVar("I", bound=SampleBase, covariant=True)


class _DataLoaderKwargs(TypedDict, total=False):
    """Dictionary of keyword arguments used to avoid specifying the default values."""

    batch_size: Optional[int]
    shuffle: bool
    sampler: Optional[Sampler[int]]
    batch_sampler: Optional[Sampler[List[int]]]
    num_workers: int
    pin_memory: bool
    drop_last: bool
    timeout: float
    worker_init_fn: Optional[_worker_init_fn_t]
    multiprocessing_context: Optional[Union[BaseContext, str]]
    generator: Optional[torch.Generator]
    prefetch_factor: int
    persistent_workers: bool
    pin_memory_device: str


class CdtDataLoader(DataLoader[I]):
    def __init__(
        self,
        dataset: SizedDataset[I],
        *,
        cast_to_sample: bool = True,
        converter: Optional[Union[Type[Any], Callable]] = None,
        **kwargs: Unpack[_DataLoaderKwargs],
    ) -> None:
        # pytorch_lightning inspects the signature and if it sees `**kwargs`, it assumes that
        # the __init__ takes all arguments that DataLoader.__init__ takes, so we have to
        # manually remove "collate_fn" here in order to avoid passing it in *twice*.
        kwargs.pop("collate_fn", None)  # type: ignore
        super().__init__(
            dataset,  # type: ignore
            collate_fn=cdt_collate(cast_to_sample=cast_to_sample, converter=converter),
            **kwargs,
        )

    def __iter__(self) -> Iterator[I]:  # type: ignore
        return super().__iter__()


def check_integrity(*, filepath: Path, md5: Optional[str]) -> None:
    from torchvision.datasets.utils import check_integrity  # type: ignore

    ext = filepath.suffix
    if ext not in [".zip", ".7z"] and check_integrity(fpath=str(filepath), md5=md5):
        raise RuntimeError('Dataset corrupted; try deleting it and redownloading it.')


class UrlFileInfo(NamedTuple):
    name: str
    url: str
    md5: Optional[str] = None


def download_from_url(
    *,
    file_info: Union[UrlFileInfo, List[UrlFileInfo]],
    root: Union[Path, str],
    logger: Optional[logging.Logger] = None,
    remove_finished: bool = True,
) -> None:
    logger = logging.getLogger(__name__) if logger is None else logger
    file_info_ls = file_info if isinstance(file_info, list) else [file_info]
    if not isinstance(root, Path):
        root = Path(root).expanduser()
    # Create the specified root directory if it doesn't already exist
    root.mkdir(parents=True, exist_ok=True)

    for info in file_info_ls:
        filepath = root / info.name

        filepath_str = str(filepath)
        suffix = _detect_file_type(filepath_str)[0]
        extracted_filepath = Path(filepath_str.split(suffix)[0])

        if extracted_filepath.exists():
            logger.info(f"File '{info.name}' already downloaded and extracted.")
        else:
            if filepath.exists():
                logger.info(f"File '{info.name}' already downloaded.")
            else:
                logger.info(f"Downloading file '{info.name}' from address '{info.url}'.")
                download_url(url=info.url, filename=info.name, root=str(root), md5=info.md5)

            logger.info(f"Extracting '{filepath.resolve()}' to '{root.resolve()}'")
            try:
                extract_archive(
                    from_path=str(filepath),
                    to_path=str(extracted_filepath),
                    remove_finished=remove_finished,
                )
            # Fall back on using jar to unzip the archive
            except BadZipFile:
                try:
                    subprocess.run(["jar", "-xvf", str(filepath)], check=True, cwd=root)
                except subprocess.CalledProcessError:
                    logger.info(
                        "Attempted to fall back on using Java to extract malformed .zip file; "
                        "however, there was a problem. Try redownloading the zip file or "
                        "checking that Java has been properly added to your system variables."
                    )


class GdriveFileInfo(NamedTuple):
    name: str
    id: str
    md5: Optional[str] = None


def download_from_gdrive(
    *,
    file_info: Union[GdriveFileInfo, List[GdriveFileInfo]],
    root: Union[Path, str],
    logger: Optional[logging.Logger] = None,
) -> None:
    """Attempt to download data if files cannot be found in the root directory."""

    logger = logging.getLogger(__name__) if logger is None else logger

    file_info_ls = file_info if isinstance(file_info, list) else [file_info]
    if not isinstance(root, Path):
        root = Path(root).expanduser()
    # Create the specified root directory if it doesn't already exist
    root.mkdir(parents=True, exist_ok=True)

    for info in file_info_ls:
        filepath = root / info.name
        if filepath.exists():
            logger.info(f"File '{info.name}' already downloaded.")
        else:
            import gdown  # type: ignore

            logger.info(f"Downloading file '{info.name}' from Google Drive.")
            gdown.cached_download(
                url=f"https://drive.google.com/uc?id={info.id}",
                path=str(filepath),
                quiet=False,
                md5=info.md5,
            )
        if filepath.suffix == ".zip":
            if filepath.with_suffix("").exists():
                logger.info(f"File '{info.name}' already unzipped.")
            else:
                check_integrity(filepath=filepath, md5=info.md5)
                # ------------------------------ Unzip the data ------------------------------
                import zipfile

                logger.info(f"Unzipping '{filepath.resolve()}'; this could take a while.")
                with zipfile.ZipFile(filepath, "r") as fhandle:
                    fhandle.extractall(str(root))


@overload
def random_split(
    dataset: PseudoCdtDataset,
    *,
    props: Union[Sequence[float], float],
    deep: bool = ...,
    as_indices: Literal[True],
    seed: Optional[int] = ...,
    reproducible: bool = ...,
) -> List[List[int]]:
    ...


@overload
def random_split(
    dataset: PCD,
    *,
    props: Union[Sequence[float], float],
    deep: bool = ...,
    as_indices: Literal[False] = ...,
    seed: Optional[int] = ...,
    reproducible: bool = ...,
) -> List[PCD]:
    ...


@overload
def random_split(
    dataset: PCD,
    *,
    props: Union[Sequence[float], float],
    deep: bool = ...,
    as_indices: bool,
    seed: Optional[int] = ...,
    reproducible: bool = ...,
) -> Union[List[PCD], List[List[int]]]:
    ...


def random_split(
    dataset: PCD,
    *,
    props: Union[Sequence[float], float],
    deep: bool = False,
    as_indices: bool = False,
    seed: Optional[int] = None,
    reproducible: bool = False,
) -> Union[List[PCD], List[List[int]]]:
    """Randomly split the dataset into subsets according to the given proportions.

    :param dataset: The dataset to split.
    :param props: The fractional size of each subset into which to randomly split the data.
        Elements must be non-negative and sum to 1 or less; if less then the size of the final
        split will be computed by complement.

    :param deep: Whether to create a copy of the underlying dataset as a basis for the random
        subsets. If False then the data of the subsets will be views of original dataset's data.

    :param as_indices: Whether to return the raw train/test indices instead of subsets of the
        dataset constructed from them.

    :param seed: PRNG seed to use for splitting the data.

    :param reproducible: Whether to make the split reproducible.

    :returns: Random subsets of the data (or their associated indices) of the requested proportions.
    """
    split_indices = prop_random_split(
        dataset, props=props, as_indices=True, seed=seed, reproducible=reproducible
    )
    if as_indices:
        return split_indices
    splits = [make_subset(dataset, indices=indices, deep=deep) for indices in split_indices]
    return splits


@overload
def stratified_split(
    dataset: PseudoCdtDataset,
    *,
    default_train_prop: float,
    train_props: Optional[Mapping[int, Union[Dict[int, float], float]]] = ...,
    seed: Optional[int] = ...,
    as_indices: Literal[True],
    reproducible: bool = False,
) -> TrainTestSplit[List[int]]:
    ...


@overload
def stratified_split(
    dataset: PCD,
    *,
    default_train_prop: float,
    train_props: Optional[Mapping[int, Union[Dict[int, float], float]]] = ...,
    seed: Optional[int] = ...,
    as_indices: Literal[False] = ...,
    reproducible: bool = False,
) -> TrainTestSplit[PCD]:
    ...


@overload
def stratified_split(
    dataset: PCD,
    *,
    default_train_prop: float,
    train_props: Optional[Mapping[int, Union[Dict[int, float], float]]] = ...,
    seed: Optional[int] = ...,
    as_indices: bool,
    reproducible: bool = False,
) -> Union[TrainTestSplit[PCD], TrainTestSplit[List[int]]]:
    ...


def stratified_split(
    dataset: PCD,
    *,
    default_train_prop: float,
    train_props: Optional[Mapping[int, Union[Dict[int, float], float]]] = None,
    seed: Optional[int] = None,
    as_indices: bool = False,
    reproducible: bool = False,
) -> Union[TrainTestSplit[PCD], TrainTestSplit[List[int]]]:
    """Splits the data into train/test sets conditional on super- and sub-class labels.

    :param dataset: The dataset to split.
    :param default_train_prop: Proportion of samples for a given to sample for
        the training set for those y-s combinations not specified in ``train_props``.
    :param train_props: Proportion of each superclass-subclass combination to sample for
        the training set. Keys correspond to the superclass while values can be either a float,
        in which case the proportion is applied to the superclass as a whole, or a dict, in which
        case the sampling is applied only to the subclasses of the superclass given by the keys.
        If ``None`` then the function reduces to a simple random split of the data.
    :param as_indices: Whether to return the raw train/test indices instead of subsets of the
        dataset constructed from them.
    :param seed: PRNG seed to use for determining the splits.
    :param reproducible: Whether to make the split reproducible.
    :returns: Train-test subsets (``as_indices=False``) or indices (``as_indices=True``).
    :raises TypeError: if no superclass labels are available.
    :raises ValueError: if the labels are not as expected.
    """
    if dataset.y is None:
        raise TypeError(
            f"Dataset of type {dataset.__class__.__name__} has no superclass labels to use "
            "for stratification."
        )
    group_ids = get_group_ids(dataset)
    y_unique = dataset.y.unique()
    groups, id_counts = group_ids.unique(return_counts=True)
    card_s = None if dataset.s is None else len(dataset.s.unique())
    ncols = 1 if card_s is None else card_s
    group_train_props = dict.fromkeys(groups.tolist(), default_train_prop)

    if train_props is not None:
        for superclass, value in train_props.items():
            # Apply the same splitting proportion to the entire superclass
            if superclass not in y_unique:
                raise ValueError(
                    f"No samples belonging to superclass 'y={superclass}' exist in the dataset of "
                    f"type {dataset.__class__.__name__}."
                )
            if isinstance(value, (float, int)):
                if not 0 <= value <= 1:
                    raise ValueError(
                        "All splitting proportions speicfied in 'train_props' must be in the "
                        "range [0, 1]."
                    )
                if card_s is None:
                    group_train_props[superclass] = value
                else:
                    superclass_props = dict.fromkeys(
                        range(superclass * card_s, (superclass + 1) * card_s),
                        value,
                    )
                    group_train_props.update(superclass_props)
            # Specifying proportions at the superclass/subclass level, rather than superclass-wide
            else:
                for subclass, train_prop in value.items():
                    if not 0 <= train_prop <= 1:
                        raise ValueError(
                            "All splitting proportions specified in 'train_props' must be in the "
                            "range [0, 1]."
                        )
                    group_id = superclass * ncols + subclass
                    if group_id not in groups:
                        raise ValueError(
                            f"No samples belonging to the subset '(y={superclass}', s={subclass})' "
                            f"exist in the dataset of type {dataset.__class__.__name__}."
                        )
                    group_train_props[group_id] = train_prop

    # Shuffle the samples before sampling
    if reproducible:
        if seed is None:
            raise ValueError("Reproducible splits require a PRNG seed to be specified.")
        # MT19937 isn't the best random number generator, but it's reproducible, so we're using it.
        generator = np.random.Generator(np.random.MT19937(seed))
        indices = np.arange(len(group_ids))
        generator.shuffle(indices)  # Shuffle the indices in-place.
        perm_inds = torch.from_numpy(indices)
    else:
        # Initialise the random-number generator
        generator = torch.default_generator if seed is None else torch.Generator().manual_seed(seed)
        perm_inds = torch.randperm(len(group_ids), generator=generator)
    group_ids_perm = group_ids[perm_inds]

    sort_inds = group_ids_perm.sort(dim=0, stable=True).indices
    thresholds = cast(
        Tensor, (torch.as_tensor(tuple(group_train_props.values())) * id_counts).round().long()
    )
    thresholds = torch.stack([thresholds, id_counts], dim=-1)
    thresholds[1:] += id_counts.cumsum(0)[:-1].unsqueeze(-1)

    train_test_inds = sort_inds.tensor_split(thresholds.flatten()[:-1], dim=0)
    train_inds = perm_inds[torch.cat(train_test_inds[0::2])]
    test_inds = perm_inds[torch.cat(train_test_inds[1::2])]

    if as_indices:
        return TrainTestSplit[List[int]](train=train_inds.tolist(), test=test_inds.tolist())

    train_data = make_subset(dataset=dataset, indices=train_inds)
    test_data = make_subset(dataset=dataset, indices=test_inds)

    return TrainTestSplit[PCD](train=train_data, test=test_data)


def is_tensor_list(ls: List[Any]) -> TypeGuard[List[Tensor]]:
    return isinstance(ls[0], Tensor)
